(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 5.1'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     33127,        983]*)
(*NotebookOutlinePosition[     33862,       1008]*)
(*  CellTagsIndexPosition[     33818,       1004]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["Examples", "Section"],

Cell[CellGroupData[{

Cell["Author", "Subsubsection"],

Cell["\<\
Jos\[EAcute] Luis G\[OAcute]mez Pardo
Departamento de \[CapitalAAcute]lxebra
Universidade de Santiago de Compostela
15782 Santiago de Compostela (Spain).\
\>", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Examples and tests", "Subsubsection"],

Cell["\<\
This notebook contains some examples of the use of AES encryption and \
decryption functions and of CBC-MAC authentication,  as well as some tests. \
To use it one must previously evaluate the notebook \"Rijndael.nb\" or load \
the package \"Cryptography`AES`\" (but not both). 

Start by generating a 256-bit random key:\
\>", "Text"],

Cell[BoxData[
    \(\(k1 = RandomKey[256];\)\)], "Input"],

Cell["Let's try a text string:", "Text"],

Cell[BoxData[
    \(\(p = "\<Strong cryptography makes the world a safer place\>";\)\)], \
"Input"],

Cell["\<\
Encrypting this plaintext in CBC mode, with the key just generated looks like \
this:\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(c1 = AesTextEncrypt[p, k1]\)], "Input"],

Cell[BoxData[
    \("\.89\.19\[YDoubleDot]\.a6\[LeftGuillemet]\[Divide]A\.19O\[Times]\.0b\
\[AGrave]\[IGrave]e\.1cmM\[IHat]oXQ[\[YDoubleDot].\[CapitalIDoubleDot]\
\[CapitalARing]{L0\[EHat]OM\[EHat]C\.81\[EDoubleDot]\.9d=\[OSlash]\.9b<\[SZ]\
\.0b\.0e+\.97-\[CapitalATilde]\[Cent]\.00\[Currency](\.b9\[CapitalAE]dO\.8d;\
\.aa\[OHat]d\[OSlash]6\.7fN\.be\[CapitalIHat]V>t\.ba\.bc\.0fnt\.bcY(\.00\
\[CapitalOHat]"\)], "Output"]
}, Open  ]],

Cell["Decrypting the ciphertext one obtains:", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(AesTextDecrypt[c1, k1]\)], "Input"],

Cell[BoxData[
    \("Strong cryptography makes the world a safer place"\)], "Output"]
}, Open  ]],

Cell["\<\
Since the mode was not specified, he previous encryption was done in CBC \
mode, which is the default. Next, let's encrypt the same text with the same \
key in CFB mode:\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(c2 = AesTextEncrypt[p, k1, "\<cfb\>"]\)], "Input"],

Cell[BoxData[
    \("\.9a6za\.9a\[CapitalCCedilla]\[CCedilla]\.16\[AHat]\.86\.00}/\.13L\
\[Copyright](\[DownExclamation]'\.17\[CapitalAE]?8U\.97\[OTilde]\.9e/\.90\
\[NonBreakingSpace]Uxy\[IDoubleDot]vC\.9e\[CapitalThorn]\"G\
\[RegisteredTrademark]\[Micro]\[CapitalOTilde]\[Copyright]K\[CapitalEth]^M\
\[AGrave]\[Degree]\.15\[ATilde]\.aa\.b4\[OTilde]M\.11\[ADoubleDot]J\.10\
\[CapitalEAcute]\[CapitalUDoubleDot]v\.86\.b9\[CapitalOSlash]\.93\.96\.9bT%x1\
\.88\[RawEscape]\[Cent]\[CapitalOSlash]|\[Thorn]\.07"\)], "Output"]
}, Open  ]],

Cell["\<\
Let's check now that decryption in CFB mode works as expected:\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(AesTextDecrypt[c2, k1, "\<cfb\>"]\)], "Input"],

Cell[BoxData[
    \("Strong cryptography makes the world a safer place"\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "Observe that ECB mode can be checked in a similar way and that if we \
repeatedly encrypt the same text string with the same key (for example, ",
  StyleBox["p",
    FontFamily->"Courier New"],
  " with ",
  StyleBox["k1",
    FontFamily->"Courier New"],
  ") in CBC or CFB mode, the ciphertext obtained is different each time. In \
ECB mode, however, only the last ciphertext block will be different due to \
the random padding. Let's now produce a 128-bit random key:"
}], "Text"],

Cell[BoxData[
    \(\(k2 = RandomKey[128];\)\)], "Input"],

Cell[TextData[{
  "Next, assuming that this notebook, with filename \"Rijndael Examples.nb\" \
(on slow computers it might be advisable to choose a smaller file for this \
example) is in the current working directory, let's encrypt it in CBC mode, \
using key \"",
  StyleBox["k2",
    FontFamily->"Courier New"],
  "\" and saving the encrypted file under the name \"Encrypted Rijndael \
Examples.nb\":"
}], "Text"],

Cell[BoxData[
    \(\(AesBinaryEncrypt["\<Rijndael Examples.nb\>", "\<Encrypted Rijndael \
Examples.nb\>", k2];\)\)], "Input"],

Cell["\<\
If we open \"Encrypted Rijndael Examples.nb\" it will look like gibberish but \
now, let's decrypt it to the file \"Decrypted Rijndael Examples.nb\":\
\>", "Text"],

Cell[BoxData[
    \(\(AesBinaryDecrypt["\<Encrypted Rijndael Examples.nb\>", "\<Decrypted \
Rijndael Examples.nb\>", k2];\)\)], "Input"],

Cell[TextData[{
  "There should now exist a perfect copy of \"Rijndael Examples.nb\", called \
\"Decrypted Rijndael Examples.nb\", in the hard disk...\n\tNext, we can try \
binary encryption in ECB mode using ",
  StyleBox["the same file names as before.",
    FontFamily->"Times New Roman"],
  StyleBox[" ",
    FontFamily->"Courier New"],
  StyleBox["Since now a file called ",
    FontFamily->"Times New Roman"],
  "\"Encrypted Rijndael Examples.nb\"",
  StyleBox[" exists in the working directory, we will be prompted about \
whether this file should be overwritten:",
    FontFamily->"Times New Roman"]
}], "Text"],

Cell[BoxData[
    \(\(AesBinaryEncrypt["\<Rijndael Examples.nb\>", "\<Encrypted Rijndael \
Examples.nb\>", k2, \ "\<ecb\>"];\)\)], "Input"],

Cell[TextData[{
  "Assuming that we chose to overwrite the file, we can now decrypt again \
\"Encrypted Rijndael Examples.nb\". If we choose to save the decrypted file \
under the name \"Decrypted Rijndael Examples.nb\" as before, we will be \
prompted again about whether this file should be overwritten. We can, \
however, overwrite the file without being prompted by using the option ",
  StyleBox["CheckFile->False",
    FontFamily->"Courier New"],
  StyleBox[":",
    FontFamily->"Times New Roman"]
}], "Text"],

Cell[BoxData[
    \(\(AesBinaryDecrypt["\<Encrypted Rijndael Examples.nb\>", "\<Decrypted \
Rijndael Examples.nb\>", k2, "\<ecb\>", 
        CheckFile \[Rule] False];\)\)], "Input"],

Cell["\<\
Again, one can check that \"Decrypted Rijndael Examples.nb\" is identical to \
the original \"Rijndael Examples.nb\". Next, we will produce a MAC for a \
file. Suppose that Alice  is ordering her bank to make a payment to Eve (E). \
His message could be the following  text string:\
\>", "Text"],

Cell[BoxData[
    \(\(m1 = "\<Transfer $1000 from my account to E's account.\>";\)\)], \
"Input"],

Cell["\<\
Suppose that the message is to be authenticated by using EMAC and keys k1 and \
k2 above, which we assume are shared by Alice and the bank. First, Alice \
writes the message to a file:\
\>", "Text"],

Cell[BoxData[
    \(\(StringToFile[m1, "\<testfile1\>"];\)\)], "Input"],

Cell["\<\
Then, Alice computes the MAC that will be used to tag the message:\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(AesCbcEmac["\<testfile1\>", BytesToHexString[k1], 
      BytesToHexString[k2]]\)], "Input"],

Cell[BoxData[
    \("3f82c300f0043f7f128cb8cb9868cbb2"\)], "Output"]
}, Open  ]],

Cell["\<\
Suppose now that the message is to be authenticated by using OMAC with the \
following key which is shared by Alice and the bank:\
\>", "Text"],

Cell[BoxData[
    \(\(k3 = "\<2b7e151628aed2a6abf7158809cf4f3c\>";\)\)], "Input"],

Cell["\<\
To authenticate the message, Alice computes the MAC that will be sent \
attached to it:\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(AesTextOmac[m1, k3]\)], "Input"],

Cell[BoxData[
    \("7a509a395ebf56fdeac190487f53cdfc"\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "She can also use ",
  StyleBox["AesOmac",
    FontFamily->"Courier New"],
  StyleBox[" to compute the tag",
    FontFamily->"Times New Roman"],
  "; of course, since the message was written to the file in binary form, \
both tags will result to be identical:"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(AesOmac["\<testfile1\>", k3]\)], "Input"],

Cell[BoxData[
    \("7a509a395ebf56fdeac190487f53cdfc"\)], "Output"]
}, Open  ]],

Cell["\<\
The verification algorithm is the same used to compute the MAC, so the bank \
will verify the authenticity and integrity of the message by computing again \
the MAC for the received message and checking that it agrees with the one \
received. Now, Eve would like to change the message received by the bank as \
follows:\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(m2 = StringReplace[m1, "\<1\>" -> "\<9\>"]\)], "Input"],

Cell[BoxData[
    \("Transfer $9000 from my account to E's account."\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "Suppose that she replaces ",
  StyleBox["m1",
    FontFamily->"Courier New"],
  " by ",
  StyleBox["m2",
    FontFamily->"Courier New"],
  ". Since she does not know the secret key ",
  StyleBox["k3",
    FontFamily->"Courier New"],
  ", she is not able to compute the correct MAC for ",
  StyleBox["m2",
    FontFamily->"Courier New"],
  " which is:"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(AesTextOmac[m2, k3]\)], "Input"],

Cell[BoxData[
    \("86f9216dc43fa81c79f9b10f68d485a7"\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "Upon receiving ",
  StyleBox["m2",
    FontFamily->"Courier New"],
  ", the bank computes this MAC. If ",
  StyleBox["m2",
    FontFamily->"Courier New"],
  " arrives to the bank with the MAC corresponding to ",
  StyleBox["m1",
    FontFamily->"Courier New"],
  ", which is completely different, authentication fails and the message will \
be rejected. Observe that ",
  StyleBox["m1",
    FontFamily->"Courier New"],
  " and ",
  StyleBox["m2",
    FontFamily->"Courier New"],
  " only differ in one byte, corresponding to the first 128-bit block \
(actually, only in one bit if we regard the messages as bit strings) but the \
corresponding MAC's have all their bytes different (in fact, in similar \
situations,  all the bytes will be different most of the time, as OMAC \
behaves like a pseudorandom function assuming that Rijndael is a pseudorandom \
permutation).  The important thing, however, is not how different these MAC's \
are but, rather, the fact that it would be very difficult for Eve to figure \
out the MAC corresponding to ",
  StyleBox["m2",
    FontFamily->"Courier New"],
  " under key ",
  StyleBox["k3",
    FontFamily->"Courier New"],
  " (which is unknown to her). In fact, the security proof for OMAC implies \
that it would be very difficult for Eve to produce a forgery consisting of a \
pair formed by a message and its valid MAC under key ",
  StyleBox["k3",
    FontFamily->"Courier New"],
  ", even if the forged message does not entail a financial gain for her or \
is completely meaningless."
}], "Text"],

Cell[TextData[{
  "The two last MAC's we computed are of type OMAC1, which is the default for \
the OMAC implementation.The tag corresponding to ",
  StyleBox["m1",
    FontFamily->"Courier New"],
  ", using key ",
  StyleBox["k1",
    FontFamily->"Courier New"],
  StyleBox[", under OMAC2 is:",
    FontFamily->"Times New Roman"]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(AesTextOmac[m1, k3, 2]\)], "Input"],

Cell[BoxData[
    \("b7fd7b80cf854b429d01821112b618e6"\)], "Output"]
}, Open  ]],

Cell["\<\
Now, let us show on screen some of the Rijndael tables and constants. First, \
the S Box:\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(DisplayRijndaelObject["\<Sbox\>"]\)], "Input"],

Cell[BoxData[
    TagBox[GridBox[{
          {"\<\"\"\>", "\<\"0\"\>", "\<\"1\"\>", "\<\"2\"\>", "\<\"3\"\>", \
"\<\"4\"\>", "\<\"5\"\>", "\<\"6\"\>", "\<\"7\"\>", "\<\"8\"\>", "\<\"9\"\>", \
"\<\"a\"\>", "\<\"b\"\>", "\<\"c\"\>", "\<\"d\"\>", "\<\"e\"\>", \
"\<\"f\"\>"},
          {"\<\"0\"\>", "\<\"63\"\>", "\<\"7c\"\>", "\<\"77\"\>", \
"\<\"7b\"\>", "\<\"f2\"\>", "\<\"6b\"\>", "\<\"6f\"\>", "\<\"c5\"\>", "\<\"30\
\"\>", "\<\"01\"\>", "\<\"67\"\>", "\<\"2b\"\>", "\<\"fe\"\>", "\<\"d7\"\>", \
"\<\"ab\"\>", "\<\"76\"\>"},
          {"\<\"1\"\>", "\<\"ca\"\>", "\<\"82\"\>", "\<\"c9\"\>", \
"\<\"7d\"\>", "\<\"fa\"\>", "\<\"59\"\>", "\<\"47\"\>", "\<\"f0\"\>", "\<\"ad\
\"\>", "\<\"d4\"\>", "\<\"a2\"\>", "\<\"af\"\>", "\<\"9c\"\>", "\<\"a4\"\>", \
"\<\"72\"\>", "\<\"c0\"\>"},
          {"\<\"2\"\>", "\<\"b7\"\>", "\<\"fd\"\>", "\<\"93\"\>", \
"\<\"26\"\>", "\<\"36\"\>", "\<\"3f\"\>", "\<\"f7\"\>", "\<\"cc\"\>", "\<\"34\
\"\>", "\<\"a5\"\>", "\<\"e5\"\>", "\<\"f1\"\>", "\<\"71\"\>", "\<\"d8\"\>", \
"\<\"31\"\>", "\<\"15\"\>"},
          {"\<\"3\"\>", "\<\"04\"\>", "\<\"c7\"\>", "\<\"23\"\>", \
"\<\"c3\"\>", "\<\"18\"\>", "\<\"96\"\>", "\<\"05\"\>", "\<\"9a\"\>", "\<\"07\
\"\>", "\<\"12\"\>", "\<\"80\"\>", "\<\"e2\"\>", "\<\"eb\"\>", "\<\"27\"\>", \
"\<\"b2\"\>", "\<\"75\"\>"},
          {"\<\"4\"\>", "\<\"09\"\>", "\<\"83\"\>", "\<\"2c\"\>", \
"\<\"1a\"\>", "\<\"1b\"\>", "\<\"6e\"\>", "\<\"5a\"\>", "\<\"a0\"\>", "\<\"52\
\"\>", "\<\"3b\"\>", "\<\"d6\"\>", "\<\"b3\"\>", "\<\"29\"\>", "\<\"e3\"\>", \
"\<\"2f\"\>", "\<\"84\"\>"},
          {"\<\"5\"\>", "\<\"53\"\>", "\<\"d1\"\>", "\<\"00\"\>", \
"\<\"ed\"\>", "\<\"20\"\>", "\<\"fc\"\>", "\<\"b1\"\>", "\<\"5b\"\>", "\<\"6a\
\"\>", "\<\"cb\"\>", "\<\"be\"\>", "\<\"39\"\>", "\<\"4a\"\>", "\<\"4c\"\>", \
"\<\"58\"\>", "\<\"cf\"\>"},
          {"\<\"6\"\>", "\<\"d0\"\>", "\<\"ef\"\>", "\<\"aa\"\>", \
"\<\"fb\"\>", "\<\"43\"\>", "\<\"4d\"\>", "\<\"33\"\>", "\<\"85\"\>", "\<\"45\
\"\>", "\<\"f9\"\>", "\<\"02\"\>", "\<\"7f\"\>", "\<\"50\"\>", "\<\"3c\"\>", \
"\<\"9f\"\>", "\<\"a8\"\>"},
          {"\<\"7\"\>", "\<\"51\"\>", "\<\"a3\"\>", "\<\"40\"\>", \
"\<\"8f\"\>", "\<\"92\"\>", "\<\"9d\"\>", "\<\"38\"\>", "\<\"f5\"\>", "\<\"bc\
\"\>", "\<\"b6\"\>", "\<\"da\"\>", "\<\"21\"\>", "\<\"10\"\>", "\<\"ff\"\>", \
"\<\"f3\"\>", "\<\"d2\"\>"},
          {"\<\"8\"\>", "\<\"cd\"\>", "\<\"0c\"\>", "\<\"13\"\>", \
"\<\"ec\"\>", "\<\"5f\"\>", "\<\"97\"\>", "\<\"44\"\>", "\<\"17\"\>", "\<\"c4\
\"\>", "\<\"a7\"\>", "\<\"7e\"\>", "\<\"3d\"\>", "\<\"64\"\>", "\<\"5d\"\>", \
"\<\"19\"\>", "\<\"73\"\>"},
          {"\<\"9\"\>", "\<\"60\"\>", "\<\"81\"\>", "\<\"4f\"\>", \
"\<\"dc\"\>", "\<\"22\"\>", "\<\"2a\"\>", "\<\"90\"\>", "\<\"88\"\>", "\<\"46\
\"\>", "\<\"ee\"\>", "\<\"b8\"\>", "\<\"14\"\>", "\<\"de\"\>", "\<\"5e\"\>", \
"\<\"0b\"\>", "\<\"db\"\>"},
          {"\<\"a\"\>", "\<\"e0\"\>", "\<\"32\"\>", "\<\"3a\"\>", \
"\<\"0a\"\>", "\<\"49\"\>", "\<\"06\"\>", "\<\"24\"\>", "\<\"5c\"\>", "\<\"c2\
\"\>", "\<\"d3\"\>", "\<\"ac\"\>", "\<\"62\"\>", "\<\"91\"\>", "\<\"95\"\>", \
"\<\"e4\"\>", "\<\"79\"\>"},
          {"\<\"b\"\>", "\<\"e7\"\>", "\<\"c8\"\>", "\<\"37\"\>", \
"\<\"6d\"\>", "\<\"8d\"\>", "\<\"d5\"\>", "\<\"4e\"\>", "\<\"a9\"\>", "\<\"6c\
\"\>", "\<\"56\"\>", "\<\"f4\"\>", "\<\"ea\"\>", "\<\"65\"\>", "\<\"7a\"\>", \
"\<\"ae\"\>", "\<\"08\"\>"},
          {"\<\"c\"\>", "\<\"ba\"\>", "\<\"78\"\>", "\<\"25\"\>", \
"\<\"2e\"\>", "\<\"1c\"\>", "\<\"a6\"\>", "\<\"b4\"\>", "\<\"c6\"\>", "\<\"e8\
\"\>", "\<\"dd\"\>", "\<\"74\"\>", "\<\"1f\"\>", "\<\"4b\"\>", "\<\"bd\"\>", \
"\<\"8b\"\>", "\<\"8a\"\>"},
          {"\<\"d\"\>", "\<\"70\"\>", "\<\"3e\"\>", "\<\"b5\"\>", \
"\<\"66\"\>", "\<\"48\"\>", "\<\"03\"\>", "\<\"f6\"\>", "\<\"0e\"\>", "\<\"61\
\"\>", "\<\"35\"\>", "\<\"57\"\>", "\<\"b9\"\>", "\<\"86\"\>", "\<\"c1\"\>", \
"\<\"1d\"\>", "\<\"9e\"\>"},
          {"\<\"e\"\>", "\<\"e1\"\>", "\<\"f8\"\>", "\<\"98\"\>", \
"\<\"11\"\>", "\<\"69\"\>", "\<\"d9\"\>", "\<\"8e\"\>", "\<\"94\"\>", "\<\"9b\
\"\>", "\<\"1e\"\>", "\<\"87\"\>", "\<\"e9\"\>", "\<\"ce\"\>", "\<\"55\"\>", \
"\<\"28\"\>", "\<\"df\"\>"},
          {"\<\"f\"\>", "\<\"8c\"\>", "\<\"a1\"\>", "\<\"89\"\>", \
"\<\"0d\"\>", "\<\"bf\"\>", "\<\"e6\"\>", "\<\"42\"\>", "\<\"68\"\>", "\<\"41\
\"\>", "\<\"99\"\>", "\<\"2d\"\>", "\<\"0f\"\>", "\<\"b0\"\>", "\<\"54\"\>", \
"\<\"bb\"\>", "\<\"16\"\>"}
          },
        RowSpacings->1,
        ColumnSpacings->1,
        RowAlignments->Baseline,
        ColumnAlignments->{Left}],
      Function[ BoxForm`e$, 
        TableForm[ 
        BoxForm`e$, TableHeadings -> {{"0", "1", "2", "3", "4", "5", "6", "7",
           "8", "9", "a", "b", "c", "d", "e", "f"}, {"0", "1", "2", "3", "4", 
          "5", "6", "7", "8", "9", "a", "b", "c", "d", "e", "f"}}, 
          TableSpacing -> {1, 1}]]]], "Output"]
}, Open  ]],

Cell[TextData[{
  "Next, the matrix used in ",
  StyleBox["InvMixColumns",
    FontFamily->"Courier New"],
  ":"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(DisplayRijndaelObject["\<IMC\>"]\)], "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {"\<\"0e\"\>", "\<\"0b\"\>", "\<\"0d\"\>", "\<\"09\"\>"},
            {"\<\"09\"\>", "\<\"0e\"\>", "\<\"0b\"\>", "\<\"0d\"\>"},
            {"\<\"0d\"\>", "\<\"09\"\>", "\<\"0e\"\>", "\<\"0b\"\>"},
            {"\<\"0b\"\>", "\<\"0d\"\>", "\<\"09\"\>", "\<\"0e\"\>"}
            }], "\[NoBreak]", ")"}],
      Function[ BoxForm`e$, 
        MatrixForm[ BoxForm`e$]]]], "Output"]
}, Open  ]],

Cell[TextData[{
  "Finally, the round constants, ",
  StyleBox["Rcon",
    FontFamily->"Courier New"],
  ":"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(DisplayRijndaelObject["\<Rcon\>"]\)], "Input"],

Cell[BoxData[
    TagBox[GridBox[{
          {"1", "\<\"01\"\>", "\<\"00\"\>", "\<\"00\"\>", "\<\"00\"\>"},
          {"2", "\<\"02\"\>", "\<\"00\"\>", "\<\"00\"\>", "\<\"00\"\>"},
          {"3", "\<\"04\"\>", "\<\"00\"\>", "\<\"00\"\>", "\<\"00\"\>"},
          {"4", "\<\"08\"\>", "\<\"00\"\>", "\<\"00\"\>", "\<\"00\"\>"},
          {"5", "\<\"10\"\>", "\<\"00\"\>", "\<\"00\"\>", "\<\"00\"\>"},
          {"6", "\<\"20\"\>", "\<\"00\"\>", "\<\"00\"\>", "\<\"00\"\>"},
          {"7", "\<\"40\"\>", "\<\"00\"\>", "\<\"00\"\>", "\<\"00\"\>"},
          {"8", "\<\"80\"\>", "\<\"00\"\>", "\<\"00\"\>", "\<\"00\"\>"},
          {"9", "\<\"1b\"\>", "\<\"00\"\>", "\<\"00\"\>", "\<\"00\"\>"},
          {"10", "\<\"36\"\>", "\<\"00\"\>", "\<\"00\"\>", "\<\"00\"\>"},
          {"11", "\<\"6c\"\>", "\<\"00\"\>", "\<\"00\"\>", "\<\"00\"\>"},
          {"12", "\<\"d8\"\>", "\<\"00\"\>", "\<\"00\"\>", "\<\"00\"\>"},
          {"13", "\<\"ab\"\>", "\<\"00\"\>", "\<\"00\"\>", "\<\"00\"\>"},
          {"14", "\<\"4d\"\>", "\<\"00\"\>", "\<\"00\"\>", "\<\"00\"\>"}
          },
        RowSpacings->1,
        ColumnSpacings->3,
        RowAlignments->Baseline,
        ColumnAlignments->{Left}],
      Function[ BoxForm`e$, 
        TableForm[ 
        BoxForm`e$, TableHeadings -> {Automatic, None}]]]], "Output"]
}, Open  ]],

Cell["\<\
Next, an example of the use of the KAT's (\"Known Answer Tests\") defined by \
NIST. Let's run the variable text tests for a 192-bit keylength (due to its \
length we do not include the printed output here):\
\>", "Text"],

Cell[BoxData[
    \(VarKeyTest[192]\)], "Input"],

Cell["\<\
Similarly, we can, for example, run  VarTextTest with a 256-bit key:\
\>", "Text"],

Cell[BoxData[
    \(VarTextTest[256]\)], "Input"],

Cell[TextData[{
  "To test the CBC and CFB modes, we will use some of the example vectors \
given in [M. Dworkin, Recommendation for Block Cipher Modes of Operation, \
Methods and Techniques, ",
  ButtonBox["http://csrc.nist.gov/publications/nistpubs/800-38a/sp800-38a.pdf \
",
    ButtonData:>{
      URL[ 
      "http://csrc.nist.gov/publications/nistpubs/800-38a/sp800-38a.pdf"], 
      None},
    ButtonStyle->"Hyperlink"],
  "]. The following plaintext, keys, and IV (and also key ",
  StyleBox["k3",
    FontFamily->"Courier New"],
  " already defined), given as hex strings,  are taken from this reference \
(note that this IV is exactly the same as the one taken by default in \
EncryptTest and DecryptTest, so that we can omit the IV in this case):"
}], "Text"],

Cell[BoxData[{
    \(\(pl = \
"\<6bc1bee22e409f96e93d7e117393172aae2d8a571e03ac9c9eb76fac45af8e5130c81c46a35\
ce411e5fbc1191a0a52eff69f2445df4f9b17ad2b417be66c3710\>";\)\), "\
\[IndentingNewLine]", 
    \(\(k4 = "\<8e73b0f7da0e6452c810f32b809079e562f8ead2522c6b7b\>";\)\), "\
\[IndentingNewLine]", 
    \(\(k5 = \
"\<603deb1015ca71be2b73aef0857d77811f352c073b6108d72d9810a30914dff4\>";\)\), \
"\[IndentingNewLine]", 
    \(\(iv = BytesToHexString[Range[16] - 1];\)\)}], "Input"],

Cell[TextData[{
  "We can also display the plaintext ",
  StyleBox["pl",
    FontFamily->"Courier New"],
  " as a table of four hex strings, one corresponding to each plaintext block \
by using, for example, the following function:"
}], "Text"],

Cell[BoxData[
    \(\(DisplayBlocks[str_String] := 
        TableForm[\(BytesToHexString[#] &\) /@ 
            Partition[HexStringToBytes[str], 16]];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(DisplayBlocks[pl]\)], "Input"],

Cell[BoxData[
    InterpretationBox[GridBox[{
          {"\<\"6bc1bee22e409f96e93d7e117393172a\"\>"},
          {"\<\"ae2d8a571e03ac9c9eb76fac45af8e51\"\>"},
          {"\<\"30c81c46a35ce411e5fbc1191a0a52ef\"\>"},
          {"\<\"f69f2445df4f9b17ad2b417be66c3710\"\>"}
          },
        RowSpacings->1,
        ColumnSpacings->3,
        RowAlignments->Baseline,
        ColumnAlignments->{Left}],
      TableForm[ {"6bc1bee22e409f96e93d7e117393172a", 
        "ae2d8a571e03ac9c9eb76fac45af8e51", 
        "30c81c46a35ce411e5fbc1191a0a52ef", 
        "f69f2445df4f9b17ad2b417be66c3710"}]]], "Output"]
}, Open  ]],

Cell[TextData[{
  "Now, let's encrypt plaintext ",
  StyleBox["pl",
    FontFamily->"Courier New"],
  " in CBC mode, using ",
  StyleBox["k3",
    FontFamily->"Courier New"],
  " as key and ",
  StyleBox["iv",
    FontFamily->"Courier New"],
  " as IV:"
}], "Text"],

Cell[BoxData[
    \(\(c = AesEncryptTest[pl, k3, "\<cbc\>"];\)\)], "Input"],

Cell[TextData[{
  "The resulting ciphertext can be compared with the one appearing in ",
  ButtonBox["[Dworkin]",
    ButtonData:>{
      URL[ 
      "http://csrc.nist.gov/publications/nistpubs/800-38a/sp800-38a.pdf"], 
      None},
    ButtonStyle->"Hyperlink"],
  ". For better readability we display it in block form:"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(DisplayBlocks[c]\)], "Input"],

Cell[BoxData[
    InterpretationBox[GridBox[{
          {"\<\"7649abac8119b246cee98e9b12e9197d\"\>"},
          {"\<\"5086cb9b507219ee95db113a917678b2\"\>"},
          {"\<\"73bed6b8e3c1743b7116e69e22229516\"\>"},
          {"\<\"3ff1caa1681fac09120eca307586e1a7\"\>"}
          },
        RowSpacings->1,
        ColumnSpacings->3,
        RowAlignments->Baseline,
        ColumnAlignments->{Left}],
      TableForm[ {"7649abac8119b246cee98e9b12e9197d", 
        "5086cb9b507219ee95db113a917678b2", 
        "73bed6b8e3c1743b7116e69e22229516", 
        "3ff1caa1681fac09120eca307586e1a7"}]]], "Output"]
}, Open  ]],

Cell[TextData[{
  "Let's now decrypt ciphertext ",
  StyleBox["c",
    FontFamily->"Courier New"],
  " to recover the original plaintext:"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(DisplayBlocks[AesDecryptTest[c, k3, "\<cbc\>"]]\)], "Input"],

Cell[BoxData[
    InterpretationBox[GridBox[{
          {"\<\"6bc1bee22e409f96e93d7e117393172a\"\>"},
          {"\<\"ae2d8a571e03ac9c9eb76fac45af8e51\"\>"},
          {"\<\"30c81c46a35ce411e5fbc1191a0a52ef\"\>"},
          {"\<\"f69f2445df4f9b17ad2b417be66c3710\"\>"}
          },
        RowSpacings->1,
        ColumnSpacings->3,
        RowAlignments->Baseline,
        ColumnAlignments->{Left}],
      TableForm[ {"6bc1bee22e409f96e93d7e117393172a", 
        "ae2d8a571e03ac9c9eb76fac45af8e51", 
        "30c81c46a35ce411e5fbc1191a0a52ef", 
        "f69f2445df4f9b17ad2b417be66c3710"}]]], "Output"]
}, Open  ]],

Cell["\<\
As another example, let's perform the encryption tests in CFB mode for the \
given 192 and 256-bit keys. The results are the following:\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(DisplayBlocks[AesEncryptTest[pl, k4, "\<cfb\>"]]\)], "Input"],

Cell[BoxData[
    InterpretationBox[GridBox[{
          {"\<\"cdc80d6fddf18cab34c25909c99a4174\"\>"},
          {"\<\"67ce7f7f81173621961a2b70171d3d7a\"\>"},
          {"\<\"2e1e8a1dd59b88b1c8e60fed1efac4c9\"\>"},
          {"\<\"c05f9f9ca9834fa042ae8fba584b09ff\"\>"}
          },
        RowSpacings->1,
        ColumnSpacings->3,
        RowAlignments->Baseline,
        ColumnAlignments->{Left}],
      TableForm[ {"cdc80d6fddf18cab34c25909c99a4174", 
        "67ce7f7f81173621961a2b70171d3d7a", 
        "2e1e8a1dd59b88b1c8e60fed1efac4c9", 
        "c05f9f9ca9834fa042ae8fba584b09ff"}]]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(DisplayBlocks[AesEncryptTest[pl, k5, "\<cfb\>"]]\)], "Input"],

Cell[BoxData[
    InterpretationBox[GridBox[{
          {"\<\"dc7e84bfda79164b7ecd8486985d3860\"\>"},
          {"\<\"39ffed143b28b1c832113c6331e5407b\"\>"},
          {"\<\"df10132415e54b92a13ed0a8267ae2f9\"\>"},
          {"\<\"75a385741ab9cef82031623d55b1e471\"\>"}
          },
        RowSpacings->1,
        ColumnSpacings->3,
        RowAlignments->Baseline,
        ColumnAlignments->{Left}],
      TableForm[ {"dc7e84bfda79164b7ecd8486985d3860", 
        "39ffed143b28b1c832113c6331e5407b", 
        "df10132415e54b92a13ed0a8267ae2f9", 
        "75a385741ab9cef82031623d55b1e471"}]]], "Output"]
}, Open  ]],

Cell["\<\
Next, let's try 512 random tests consisting of the encryption and the \
decryption of 512 random text strings of lengths varying between 1 and 64:\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(AesRandomTest[512, 64]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\(100\[InvisibleSpace]" tests"\),
      SequenceForm[ 100, " tests"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(200\[InvisibleSpace]" tests"\),
      SequenceForm[ 200, " tests"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(300\[InvisibleSpace]" tests"\),
      SequenceForm[ 300, " tests"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(400\[InvisibleSpace]" tests"\),
      SequenceForm[ 400, " tests"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(500\[InvisibleSpace]" tests"\),
      SequenceForm[ 500, " tests"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(512\[InvisibleSpace]" tests successfully passed"\),
      SequenceForm[ 512, " tests successfully passed"],
      Editable->False]], "Print"]
}, Open  ]],

Cell[TextData[{
  "Finally, let's test the binary input-output functions. We create a file \
(called ",
  StyleBox["\"testfile2\"",
    FontFamily->"Courier New"],
  ") in the current directory, containing the text string \
\"\[DownExclamation]La criptograf\[IAcute]a fuerte hace que el mundo sea m\
\[AAcute]s seguro!\", which includes some characters not belonging to the \
English alphabet but, for this test, we could use any binary file as well:"
}], "Text"],

Cell[BoxData[
    \(\(StringToFile["\<\[DownExclamation]La criptograf\[IAcute]a fuerte hace \
que el mundo sea m\[AAcute]s seguro!\>", \ "\<testfile2\>"];\)\)], "Input"],

Cell["\<\
Now, we submit this file to 100 encryption-decryption processes with random \
keys in CFB mode, overwriting \"testfile2\" after each encryption:\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(AesRandomBinaryTest["\<testfile2\>", "\<cfb\>", 100]\)], "Input"],

Cell[BoxData[
    \("Test successfully passed"\)], "Print"]
}, Open  ]],

Cell[TextData[{
  "Let's check again that ",
  StyleBox["\"testfile2\"",
    FontFamily->"Courier New"],
  " has not changed:"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(FileToString["\<testfile2\>"]\)], "Input"],

Cell[BoxData[
    \("\[DownExclamation]La criptograf\[IAcute]a fuerte hace que el mundo sea \
m\[AAcute]s seguro!"\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "Finally, we are going to test the implementation of OMAC1 and OMAC2, using \
the test vectors given by the authors in the OMAC page ",
  ButtonBox["http://crypt.cis.ibaraki.ac.jp/omac/omac.html",
    ButtonData:>{
      URL[ "http://crypt.cis.ibaraki.ac.jp/omac/omac.html"], None},
    ButtonStyle->"Hyperlink"],
  ". There are in total 12 test vectors for OMAC1 and another 12 tests \
vectors for OMAC2. In each case, these vectors arise from taking as message \
either the first 0 bytes (i.e., the empty string), the first 16 bytes, the \
first 40 bytes or the first 64 bytes of the message ",
  StyleBox["pl",
    FontFamily->"Courier New"],
  " above defined. For each of these four messages, keys k3, k4 and k5 above \
are used, giving a total of 12 MAC's for each of the two OMAC variants. Let's \
try, for example, the empty string with key k4 in OMAC1:"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(OmacTest["\<\>", k4]\)], "Input"],

Cell[BoxData[
    \("d17ddf46adaacde531cac483de7a9367"\)], "Output"]
}, Open  ]],

Cell["Next, we make a list with the four messages mentioned above:", "Text"],

Cell[BoxData[
    \(\(ms = \(StringTake[pl, #] &\) /@ {0, 32, 80, 128};\)\)], "Input"],

Cell[TextData[{
  "Now, we compute the four tags of these messages with the 128-bit key ",
  StyleBox["k3",
    FontFamily->"Courier New"],
  StyleBox[" and show them in table form",
    FontFamily->"Times New Roman"],
  StyleBox[":",
    FontFamily->"Times New Roman"]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(OmacTest[#, k3] &\) /@ ms // TableForm\)], "Input"],

Cell[BoxData[
    InterpretationBox[GridBox[{
          {"\<\"bb1d6929e95937287fa37d129b756746\"\>"},
          {"\<\"070a16b46b4d4144f79bdd9dd04a287c\"\>"},
          {"\<\"dfa66747de9ae63030ca32611497c827\"\>"},
          {"\<\"51f0bebf7e3b9d92fc49741779363cfe\"\>"}
          },
        RowSpacings->1,
        ColumnSpacings->3,
        RowAlignments->Baseline,
        ColumnAlignments->{Left}],
      TableForm[ {"bb1d6929e95937287fa37d129b756746", 
        "070a16b46b4d4144f79bdd9dd04a287c", 
        "dfa66747de9ae63030ca32611497c827", 
        "51f0bebf7e3b9d92fc49741779363cfe"}]]], "Output"]
}, Open  ]],

Cell[TextData[{
  "Similarly, the tags corresponding to the 256-bit key ",
  StyleBox["k5",
    FontFamily->"Courier New"],
  ":"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(OmacTest[#, k5] &\) /@ ms // TableForm\)], "Input"],

Cell[BoxData[
    InterpretationBox[GridBox[{
          {"\<\"028962f61b7bf89efc6b551f4667d983\"\>"},
          {"\<\"28a7023f452e8f82bd4bf28d8c37c35c\"\>"},
          {"\<\"aaf3d8f1de5640c232f5b169b9c911e6\"\>"},
          {"\<\"e1992190549f6ed5696a2c056c315410\"\>"}
          },
        RowSpacings->1,
        ColumnSpacings->3,
        RowAlignments->Baseline,
        ColumnAlignments->{Left}],
      TableForm[ {"028962f61b7bf89efc6b551f4667d983", 
        "28a7023f452e8f82bd4bf28d8c37c35c", 
        "aaf3d8f1de5640c232f5b169b9c911e6", 
        "e1992190549f6ed5696a2c056c315410"}]]], "Output"]
}, Open  ]],

Cell[TextData[{
  "And, finally, the tags corresponding to the four messages with the 192-bit \
key ",
  StyleBox["k4",
    FontFamily->"Courier New"],
  " and OMAC2:"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(OmacTest[#, k4, 2] &\) /@ ms // TableForm\)], "Input"],

Cell[BoxData[
    InterpretationBox[GridBox[{
          {"\<\"149f579df2129d45a69266898f55aeb2\"\>"},
          {"\<\"9e99a7bf31e710900662f65e617c5184\"\>"},
          {"\<\"b35e2d1b73aed49b78bdbdfe61f646df\"\>"},
          {"\<\"a1d5df0eed790f794d77589659f39a11\"\>"}
          },
        RowSpacings->1,
        ColumnSpacings->3,
        RowAlignments->Baseline,
        ColumnAlignments->{Left}],
      TableForm[ {"149f579df2129d45a69266898f55aeb2", 
        "9e99a7bf31e710900662f65e617c5184", 
        "b35e2d1b73aed49b78bdbdfe61f646df", 
        "a1d5df0eed790f794d77589659f39a11"}]]], "Output"]
}, Open  ]]
}, Closed]]
}, Closed]]
},
FrontEndVersion->"5.1 for Microsoft Windows",
ScreenRectangle->{{0, 1280}, {0, 951}},
WindowSize->{951, 512},
WindowMargins->{{Automatic, 59}, {Automatic, 187}},
PrintingCopies->1,
PrintingPageRange->{Automatic, Automatic},
CellLabelAutoDelete->True
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 27, 0, 73, "Section"],

Cell[CellGroupData[{
Cell[1828, 57, 31, 0, 29, "Subsubsection"],
Cell[1862, 59, 177, 5, 90, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[2076, 69, 43, 0, 23, "Subsubsection"],
Cell[2122, 71, 345, 7, 90, "Text"],
Cell[2470, 80, 57, 1, 30, "Input"],
Cell[2530, 83, 40, 0, 33, "Text"],
Cell[2573, 85, 99, 2, 30, "Input"],
Cell[2675, 89, 109, 3, 33, "Text"],

Cell[CellGroupData[{
Cell[2809, 96, 59, 1, 30, "Input"],
Cell[2871, 99, 422, 6, 29, "Output"]
}, Open  ]],
Cell[3308, 108, 54, 0, 33, "Text"],

Cell[CellGroupData[{
Cell[3387, 112, 55, 1, 30, "Input"],
Cell[3445, 115, 85, 1, 29, "Output"]
}, Open  ]],
Cell[3545, 119, 193, 4, 33, "Text"],

Cell[CellGroupData[{
Cell[3763, 127, 70, 1, 30, "Input"],
Cell[3836, 130, 520, 7, 29, "Output"]
}, Open  ]],
Cell[4371, 140, 86, 2, 33, "Text"],

Cell[CellGroupData[{
Cell[4482, 146, 66, 1, 30, "Input"],
Cell[4551, 149, 85, 1, 29, "Output"]
}, Open  ]],
Cell[4651, 153, 501, 11, 71, "Text"],
Cell[5155, 166, 57, 1, 30, "Input"],
Cell[5215, 169, 415, 9, 52, "Text"],
Cell[5633, 180, 126, 2, 30, "Input"],
Cell[5762, 184, 173, 3, 33, "Text"],
Cell[5938, 189, 136, 2, 30, "Input"],
Cell[6077, 193, 619, 14, 71, "Text"],
Cell[6699, 209, 139, 2, 30, "Input"],
Cell[6841, 213, 515, 10, 71, "Text"],
Cell[7359, 225, 181, 3, 30, "Input"],
Cell[7543, 230, 305, 5, 52, "Text"],
Cell[7851, 237, 97, 2, 30, "Input"],
Cell[7951, 241, 208, 4, 52, "Text"],
Cell[8162, 247, 71, 1, 30, "Input"],
Cell[8236, 250, 90, 2, 33, "Text"],

Cell[CellGroupData[{
Cell[8351, 256, 111, 2, 30, "Input"],
Cell[8465, 260, 68, 1, 29, "Output"]
}, Open  ]],
Cell[8548, 264, 153, 3, 33, "Text"],
Cell[8704, 269, 81, 1, 30, "Input"],
Cell[8788, 272, 111, 3, 33, "Text"],

Cell[CellGroupData[{
Cell[8924, 279, 52, 1, 30, "Input"],
Cell[8979, 282, 68, 1, 29, "Output"]
}, Open  ]],
Cell[9062, 286, 290, 8, 33, "Text"],

Cell[CellGroupData[{
Cell[9377, 298, 61, 1, 30, "Input"],
Cell[9441, 301, 68, 1, 29, "Output"]
}, Open  ]],
Cell[9524, 305, 343, 6, 52, "Text"],

Cell[CellGroupData[{
Cell[9892, 315, 75, 1, 30, "Input"],
Cell[9970, 318, 82, 1, 29, "Output"]
}, Open  ]],
Cell[10067, 322, 382, 14, 33, "Text"],

Cell[CellGroupData[{
Cell[10474, 340, 52, 1, 30, "Input"],
Cell[10529, 343, 68, 1, 29, "Output"]
}, Open  ]],
Cell[10612, 347, 1560, 37, 147, "Text"],
Cell[12175, 386, 342, 10, 33, "Text"],

Cell[CellGroupData[{
Cell[12542, 400, 55, 1, 30, "Input"],
Cell[12600, 403, 68, 1, 29, "Output"]
}, Open  ]],
Cell[12683, 407, 113, 3, 33, "Text"],

Cell[CellGroupData[{
Cell[12821, 414, 66, 1, 30, "Input"],
Cell[12890, 417, 4806, 80, 281, "Output"]
}, Open  ]],
Cell[17711, 500, 124, 5, 33, "Text"],

Cell[CellGroupData[{
Cell[17860, 509, 65, 1, 30, "Input"],
Cell[17928, 512, 459, 9, 73, "Output"]
}, Open  ]],
Cell[18402, 524, 120, 5, 33, "Text"],

Cell[CellGroupData[{
Cell[18547, 533, 66, 1, 30, "Input"],
Cell[18616, 536, 1311, 23, 233, "Output"]
}, Open  ]],
Cell[19942, 562, 231, 4, 52, "Text"],
Cell[20176, 568, 48, 1, 30, "Input"],
Cell[20227, 571, 92, 2, 33, "Text"],
Cell[20322, 575, 49, 1, 30, "Input"],
Cell[20374, 578, 770, 17, 71, "Text"],
Cell[21147, 597, 478, 10, 150, "Input"],
Cell[21628, 609, 244, 6, 33, "Text"],
Cell[21875, 617, 165, 3, 30, "Input"],

Cell[CellGroupData[{
Cell[22065, 624, 50, 1, 30, "Input"],
Cell[22118, 627, 603, 14, 73, "Output"]
}, Open  ]],
Cell[22736, 644, 265, 11, 33, "Text"],
Cell[23004, 657, 75, 1, 30, "Input"],
Cell[23082, 660, 333, 9, 33, "Text"],

Cell[CellGroupData[{
Cell[23440, 673, 49, 1, 30, "Input"],
Cell[23492, 676, 603, 14, 73, "Output"]
}, Open  ]],
Cell[24110, 693, 150, 5, 33, "Text"],

Cell[CellGroupData[{
Cell[24285, 702, 80, 1, 30, "Input"],
Cell[24368, 705, 603, 14, 73, "Output"]
}, Open  ]],
Cell[24986, 722, 159, 3, 33, "Text"],

Cell[CellGroupData[{
Cell[25170, 729, 81, 1, 30, "Input"],
Cell[25254, 732, 603, 14, 73, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[25894, 751, 81, 1, 30, "Input"],
Cell[25978, 754, 603, 14, 73, "Output"]
}, Open  ]],
Cell[26596, 771, 170, 3, 33, "Text"],

Cell[CellGroupData[{
Cell[26791, 778, 55, 1, 30, "Input"],
Cell[26849, 781, 139, 3, 25, "Print"],
Cell[26991, 786, 139, 3, 25, "Print"],
Cell[27133, 791, 139, 3, 25, "Print"],
Cell[27275, 796, 139, 3, 25, "Print"],
Cell[27417, 801, 139, 3, 25, "Print"],
Cell[27559, 806, 179, 3, 25, "Print"]
}, Open  ]],
Cell[27753, 812, 463, 9, 52, "Text"],
Cell[28219, 823, 169, 2, 30, "Input"],
Cell[28391, 827, 168, 3, 33, "Text"],

Cell[CellGroupData[{
Cell[28584, 834, 85, 1, 30, "Input"],
Cell[28672, 837, 59, 1, 25, "Print"]
}, Open  ]],
Cell[28746, 841, 138, 5, 33, "Text"],

Cell[CellGroupData[{
Cell[28909, 850, 62, 1, 30, "Input"],
Cell[28974, 853, 128, 2, 29, "Output"]
}, Open  ]],
Cell[29117, 858, 892, 16, 109, "Text"],

Cell[CellGroupData[{
Cell[30034, 878, 53, 1, 30, "Input"],
Cell[30090, 881, 68, 1, 29, "Output"]
}, Open  ]],
Cell[30173, 885, 76, 0, 33, "Text"],
Cell[30252, 887, 86, 1, 30, "Input"],
Cell[30341, 890, 281, 8, 33, "Text"],

Cell[CellGroupData[{
Cell[30647, 902, 73, 1, 30, "Input"],
Cell[30723, 905, 603, 14, 73, "Output"]
}, Open  ]],
Cell[31341, 922, 141, 5, 33, "Text"],

Cell[CellGroupData[{
Cell[31507, 931, 73, 1, 30, "Input"],
Cell[31583, 934, 603, 14, 73, "Output"]
}, Open  ]],
Cell[32201, 951, 179, 6, 33, "Text"],

Cell[CellGroupData[{
Cell[32405, 961, 76, 1, 30, "Input"],
Cell[32484, 964, 603, 14, 73, "Output"]
}, Open  ]]
}, Closed]]
}, Closed]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

