(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 5.1'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    128190,       3210]*)
(*NotebookOutlinePosition[    129067,       3242]*)
(*  CellTagsIndexPosition[    128995,       3236]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["The Advanced Encryption Standard", "Title"],

Cell[CellGroupData[{

Cell["Author", "Section"],

Cell[TextData[{
  StyleBox["Jos\[EAcute] Luis G\[OAcute]mez Pardo",
    FontFamily->"Arial",
    FontWeight->"Plain"],
  "\n",
  StyleBox["Departamento de Alxebra\nUniversidade de Santiago de Compostela\n\
15782 Santiago de Compostela, Spain\ne-mail: pardo@usc.es",
    FontFamily->"Arial",
    FontSize->12,
    FontWeight->"Plain"]
}], "Text",
  FontFamily->"Helvetica-Light",
  FontSize->16,
  FontWeight->"Bold"]
}, Closed]],

Cell[CellGroupData[{

Cell["Copyright Notice", "Section"],

Cell[TextData[{
  "Copyright \[Copyright] 2003-2005 Jos\[EAcute] Luis G\[OAcute]mez Pardo\n\
This notebook contains a ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " implementation of the Advanced Encryption Standard (AES) in ECB, CBC, and \
CFB128 modes, as well as several variants of AES-based CBC-MAC \
authentication, including OMAC.\nPermission is hereby granted to make copies \
of this file for any purpose other than direct profit, or as part of a \
commercial product, provided this copyright notice is left intact. Also, \
permission is granted to reproduce part or all of this file provided that the \
source is acknowledged."
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Version", "Section"],

Cell["Version 2.1", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Requirements", "Section"],

Cell[TextData[{
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  "  v. 4.0 or higher"
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Introduction", "Section"],

Cell[TextData[{
  "The purpose of this notebook is to give a brief overview of the AES \
implementation contained in it but it is not intended as an explanation of \
AES itself; for this see the references. The Rijndael algorithm, a symmetric \
block cipher with three possible key lengths (128, 192, and 256-bit), \
designed by Joan Daemen and Vincent Rijmen, was adopted as a standard by NIST \
(as the Advanced Encryption Standard or, briefly, AES)  on 2001 and also \
approved as FIPS (Federal Information Processing Standard) on the same year.  \
Originally, Rijndael had also provision for variable block size but only the \
128-bit size was retained for AES. In practice, we will not make distinction \
between AES and Rijndael and will use both names indistinctly. The original \
AES proposal ",
  ButtonBox["[3]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  " contains a very clear description of Rijndael, and the same can be said \
of the official AES specification ",
  ButtonBox["[11]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  ". Other useful descriptions are contained in the books ",
  ButtonBox["[12, 15",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  "], while ",
  ButtonBox["[4]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  " gives also a detailed account of all the relevant aspects of Rijndael's \
design. Hurried people might want to have a look at ",
  ButtonBox["[8]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  ", where a one-page description of Rijndael is given, but as the title \
already warns, some algebra knowledge is required to understand it. \
Additional interesting information on Rijndael and its implementations can be \
obtained from the web pages ",
  ButtonBox["[11, 14]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  ". The modes of operation of block symmetric ciphers are described in ",
  ButtonBox["[5]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  ", and also in  ",
  ButtonBox["[12, 15]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  ". An excellent source for understanding the role of these modes in modern \
cryptography is provided by the Bellare-Rogaway notes ",
  ButtonBox["[2]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  ". On the other hand, ",
  ButtonBox["[7]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  " gives an introduction to symmetric block ciphers in general, while ",
  ButtonBox["[16]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  " provides, in a more specialized context, a nice overview of the \
cryptanalysis of these ciphers.\n\tWhile optimizing for speed was certainly \
not the main goal in developing this code, some steps have been taken in \
order to obtain a working implementation that can actually be used to encrypt \
files of moderate size and not just short text strings. As explained in more \
detail below, some of  AES operations are based on the arithmetic of the \
field of 256 elements (more specifically, ",
  StyleBox["AddRoundKey",
    FontFamily->"Courier New"],
  " is based on the sum in this field, while ",
  StyleBox["SubBytes",
    FontFamily->"Courier New"],
  " makes use of the multiplicative inverses in the field and ",
  StyleBox["MixColumns",
    FontFamily->"Courier New"],
  " of the field multiplication; a good reference for all things related to \
finite fields is ",
  ButtonBox["[9]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  "). When the elements of this field are regarded as bytes, the sum is \
merely the bitwise XOR and so it poses no speed problems. However, the \
multiplication and the computation of multiplicative inverses usually rely on \
polynomial operations (as in the ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " package ",
  StyleBox["Algebra`FiniteFields`",
    FontFamily->"Courier New"],
  ") which are much more time-consuming and that, if used during the \
encryption and decryption processes of AES, would lead to a very slow \
implementation of the algorithm. The standard procedure to avoid this problem \
is to use lookup tables. Here, these tables are not stored as such, but are \
generated by using the field arithmetic during the initialization process, \
which is very fast on modern computers. The generation of the permutation of \
the elements of the field underlying the ",
  StyleBox["SubBytes",
    FontFamily->"Courier New"],
  " operation, called the \"",
  StyleBox["S Box",
    FontSlant->"Italic"],
  "\", is, of course, very fast, as it requires very few field operations. To \
avoid actually computing multiplications of field elements during the ",
  StyleBox["MixColumns",
    FontFamily->"Courier New"],
  " operations, a \"multiplication table\" of the field can be generated and \
used in the encryption and decryption processes as a lookup table. This can \
be done very quickly since it is not necessary to do all the multiplications, \
as the table can be computed by just XOR-ing and multiplying by the element 2 \
of the field, which is very fast. Moreover, we do not need the full \
multiplication table but just a \"reduced table\" which is sufficient for our \
purposes.  On the other hand, further optimization in speed is possible by \
constructing four lookup tables as described in the original Rijndael \
proposal ",
  ButtonBox["[3]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  ". But this would trade off coding simplicity and clarity in exchange for \
the speed gain and so, bearing in mind the educational flavor of this \
package, we decided not to make use of this possibility. The approach used \
here allows for an implementation in which no field operations (apart from \
bitwise XOR's) are ever computed  (but just looked up in tables) during the \
encryption and decryption stages, their direct use being restricted to the \
initialization process.  We feel that this provides reasonable speed within \
this context although, of course, it is still orders of magnitude slower than \
what can be achieved by using, for example, C code.\n\tThis notebook contains \
a full implementation of the ECB, CBC, and CFB128 modes of AES, which accepts \
all the keysizes allowed by the standard, namely, 128, 192, and 256-bit. The \
functional programming capabilities of ",
  StyleBox["Mathematica ",
    FontSlant->"Italic"],
  "are especially well suited for this project, as they permit, for example, \
to code the encryption and decryption functions for all the list of states \
constituting the plaintext without making explicit use of a loop for running \
through all the states. It would not be too difficult to allow for other \
modes of operation and we will consider the possibility of adding more modes \
in the future if we observe that there is sufficient interest in them. Apart \
from confidentiality, a block cipher such as AES can also provide for message \
authentication (cf.  Chapter 6 of ",
  ButtonBox["[2]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  " for a discussion of this topic) through the use of  the so-called CBC-MAC \
(see ",
  ButtonBox["[10]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  " for a description), an algorithm that produces a MAC (\"Message \
Authentication Code\") for a message by encrypting it in CBC mode with a \
public initialization vector and a shared secret key. The basic CBC-MAC \
consists of the last ciphertext block and is not secure if messages of \
variable length are used ",
  ButtonBox["[10]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  ". There are variants that offer more security in this situation and we \
implement several of them. The simpler one relies on truncating the basic MAC \
and a more secure one, called EMAC, on encrypting it again with a different \
key. The last variant of CBC-MAC which we also implement is called OMAC and \
is in the process of being standardized by NIST (see the document ",
  ButtonBox["http://csrc.nist.gov/CryptoToolkit/modes/OMAC Rationale.pdf",
    ButtonData:>{
      URL[ "http://csrc.nist.gov/CryptoToolkit/modes/OMAC Rationale.pdf"], 
      None},
    ButtonStyle->"Hyperlink"],
  "). OMAC was introduced by T. Iwata and K. Kurosawa ",
  ButtonBox["[6]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  " as a variation of the algorithm called XCBC which had been proposed \
earlier by Black and Rogaway. Like EMAC, OMAC has also been proven to be \
secure and provides better performance characteristics as it only requires \
one key scheduling (instead of two required by EMAC). In addition to \
implementing it, we give a somewhat more detailed account of the OMAC \
algorithm because of its recent origin and the fact that, as we write this, \
it has not yet appeared in books.\n\tFinally, a word of caution. Users are \
advised that the author accepts no responsibility for any problem caused by \
the use of this software. In particular, care should be taken not to lose \
data and one should always make backup copies of any files that are going to \
be encrypted and overwritten. Since Rijndael is pretty strong crypto right \
now, if a file is encrypted under the same name, without keeping a backup of \
the original,  and if the key gets lost, there is a very high probability \
that the file will be unrecoverable for some time. Of course, in general it \
is a good idea to use a different name for the encrypted version of a file, \
except perhaps when doing some testing with easily replaceable files, as in \
one of the programs included in this notebook.\n"
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Preparation", "Section"],

Cell[TextData[{
  "In this section, we set up the declaration of functions that are to be \
exported from the package \"",
  StyleBox["Rijndael`",
    FontFamily->"Courier New"],
  "\", contained in this notebook. A more complete explanation of each of \
these functions (and also of the private functions in the package) is given \
in the following sections. Also, some error messages are included."
}], "Text"],

Cell[BoxData[
    \(\(BeginPackage["\<Rijndael`\>"];\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(TextToKey::usage = \[IndentingNewLine]"\<TextToKey[text, length] \
generates a key of the specified length
   (128, 192, or 256-bit) from a text string.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(RandomKey::usage = \[IndentingNewLine]"\<RandomKey[length] generates \
a random key of the given length 
   (128, 192, or 256-bit).\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(AesTextEncrypt::usage = \[IndentingNewLine]\ \ \
"\<AesTextEncrypt[plaintext, key, (mode)] encrypts a plaintext string
    using the given key and mode. The latter can be: ecb, cbc or cfb;
    if no mode is specified, cbc is used by default.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(AesTextDecrypt::usage = \[IndentingNewLine]\ \
"\<AesTextDecrypt[ciphertext, key, (mode)] decrypts a ciphertext string
    using the given key and mode. The latter can be: ecb, cbc or cfb;
    if no mode is specified, cbc is used by default.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(AesBinaryEncrypt::usage = \[IndentingNewLine]\ \
"\<AesBinaryEncrypt[filename, cfilename, key, (mode)] reads the file named
   filename from disk, encrypts it with key in the specified mode (ecb, cbc
   or cfb), and writes it to cfilename. If the optional argument mode is
   omitted, then cbc mode is used by default. This function checks if there
   is a file named cfilename in the working directory and, in that case, the 
   user is prompted about whether this file should be overwritten or a
   different name should be used. This check can be overridden by setting the
   option CheckFile to False.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(AesBinaryDecrypt::usage = \
\[IndentingNewLine]"\<AesBinaryDecrypt[cfilename, filename, key, (mode)] \
reads the file named
   cfilename from disk, decrypts it with key in the specified mode (ecb, cbc
   or cfb), and writes it to filename. If the optional argument mode is
   omitted, then cbc mode is used by default. This function checks if there
   is a file named filename in the working directory and, in that case, the
   user is prompted about whether this file should be overwritten or a
   different name should be used. This check can be overridden by setting the
   option CheckFile to False.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(CheckFile::usage = \ \[IndentingNewLine]"\<CheckFile is an option to \
AesBinaryEncrypt and AesBinaryDecrypt. By default, 
   these functions check whether a file with the specified name exists on \
disk. If
   set to False, the file is written to disk without checking.\>";\)\)], \
"Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(StringToFile::usage = \[IndentingNewLine]"\<StringToFile[str, \
filename] writes the string str to a file named filename,
   in binary form.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(FileToString::usage = \[IndentingNewLine]"\<FileToString[filename] \
reads the file named filename in binary form and
   converts it to a string.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(BytesToHexString::usage = \[IndentingNewLine]\ \ \
"\<BytesToHexString[bytes] converts a list of bytes to a hex \
string.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(HexStringToBytes::usage = \[IndentingNewLine]"\<HexStringToBytes[str] \
converts a hex string to the corresponding list
   of bytes.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(KeyStringToBytes::usage = \
\[IndentingNewLine]"\<KeyStringToBytes[keystring] converts an AES key given \
as a hex string
   to a list of bytes.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(RandomKeyString::usage = \
\[IndentingNewLine]"\<RandomKeyString[length] generates a (pseudo-)random key \
of the specified
   length (128, 192, or 256-bit) in the form of a hex string.\>";\)\)], \
"Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(TextToKeyString::usage = \[IndentingNewLine]"\<TextToKeyString[text, \
length] generates a key of the specified length (128,
   192, or 256-bit), given as a hex string, from the supplied text.\>";\)\)], \
"Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(AesCbcMac::usage = \[IndentingNewLine]"\<AesCbcMac[filename, \
keystring] uses the secret key to compute the MAC
   associated to the file named filename. The output is an 8-byte hex string.\
\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(AesCbcEmac::usage = \[IndentingNewLine]"\<AesCbcEmac[filename, \
keystring1, keystring2] uses the secret
   keystring1 and keystring2 to compute the EMAC associated to the file
   named filename. The output is a 16-byte hex string.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(AesOmac::usage = \[IndentingNewLine]"\<AesOmac[filename, keystring, \
(type)] produces the OMAC for the given file,
   using keystring as key. The type should be given as 1 for OMAC1 and 2
   for OMAC2. If no type is specified, then OMAC1 is used by \
default.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(AesTextOmac::usage = \[IndentingNewLine]"\<AesTextOmac[text, \
keystring, (type)] produces the OMAC for the given text,
   using keystring as key. The type should be given as 1 for OMAC1 and 2
   for OMAC2. If no type is specified, then OMAC1 is used by \
default.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(DisplayRijndaelObject::usage = \
\[IndentingNewLine]"\<DisplayRijndaelObject[objectname] displays the \
corresponding Rijndael
   constant or table (objectname can be Sbox, ISbox, MC, IMC or \
Rcon).\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(VarTextTest::usage = \[IndentingNewLine]"\<VarTextTest[length] \
computes the input of the variable text tests for the
   given keylength, and produces as output the result of encrypting these \
texts.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(VarKeyTest::usage = \[IndentingNewLine]"\<VarKeyTest[length] computes \
the input of the variable key tests for the
   given keylength, and produces as output the corresponding ciphertexts.\>";\
\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(AesEncryptTest::usage = \ \
\[IndentingNewLine]"\<AesEncryptTest[plaintext, keystring, (iv), mode] \
encrypts plaintext in the
   specified mode, with the supplied key and iv (all of them given as hex
   strings). If no iv is supplied, then the string corresponding to the first
   16 bytes in ascending order is used by default (except in ECB mode, where
   it is not needed).\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(AesDecryptTest::usage = \ \
\[IndentingNewLine]"\<AesDecryptTest[ciphertext, keystring, (iv), mode] \
decrypts ciphertext in the
   specified mode, with the supplied key and iv (all of them given as hex
   strings). If no iv is supplied, then the string corresponding to the first
   16 bytes in ascending order is used by default (except in ECB mode, where \
it is
   not needed).\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(AesRandomTest::usage = \[IndentingNewLine]"\<AesRandomTest[n, l] \
generates n random text strings of length at most l and
   n random AES keys which are used to encrypt and decrypt these text strings
   For each of these tests, the encryption-decryption mode is also chosen at
   random. If all the results match, the test is declared successfully \
passed.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(AesRandomBinaryTest::usage = \
\[IndentingNewLine]"\<AesRandomBinaryTest[filename, (mode), n] submits the \
file named filename to n
   successive encryption-decryption processes in the given mode with n \
randomly
   chosen keys. If no mode is specified, then cbc is used by default. The \
test
   is successful if the final result is identical to the original file.\>";\)\
\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(OmacTest::usage = \[IndentingNewLine]"\<OmacTest[text, keystring, \
(type)] takes as input a hexadecimal text string, a
   hexadecimal keystring and the OMAC type (either 1 or 2) and produces as
   output the corresponding MAC, also in the form of a hex string. If no type
   is specified, OMAC1 is used by default.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(TextToKey::noemp = \[IndentingNewLine]"\<The empty text string cannot \
be used to generate a key.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(TextToKey::noleng = \[IndentingNewLine]"\<The key length must be \
equal to 128, 192, or 256.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(RandomKey::noleng = \[IndentingNewLine]"\<The key length must be \
equal to 128, 192, or 256.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(CheckKey::nobytes = \[IndentingNewLine]"\<The supplied key `` is not \
a list of bytes.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(CheckKey::noleng = \[IndentingNewLine]"\<The supplied key `` does not \
have valid length.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(AesTextDecrypt::notaes = \[IndentingNewLine]"\<`` is not an \
AES-encrypted text.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(AesTextDecrypt::badbyte = \[IndentingNewLine]"\<Error, probably due \
to using a wrong key or a wrong mode.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(AesBinaryDecrypt::notaes = \[IndentingNewLine]"\<`` is not an \
AES-encrypted file.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(AesBinaryDecrypt::badbyte = \[IndentingNewLine]"\<Error, probably due \
to using a wrong key or a wrong mode.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(HexStringToBytes::nohex = \[IndentingNewLine]"\<`` is not a valid hex \
string.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(KeyStringToBytes::noleng = \[IndentingNewLine]"\<The key `` does not \
have valid length.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(AesOmac::notype = \[IndentingNewLine]\ "\<The OMAC type must be \
either 1 or 2.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(DisplayRijndaelObject::noobj = \[IndentingNewLine]"\<The supplied \
argument is not a valid name of a Rijndael object.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(AesEncryptTest::noleng = \[IndentingNewLine]"\<The IV `` does not \
have valid length.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(AesDecryptTest::noleng = \[IndentingNewLine]"\<The IV `` does not \
have valid length.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(OmacTest::notype = \[IndentingNewLine]"\<The OMAC type must be either \
1 or 2.\>";\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(Begin["\<`Private`\>"];\)\)], "Input",
  InitializationCell->True]
}, Closed]],

Cell[CellGroupData[{

Cell["Field operations", "Section"],

Cell[TextData[{
  "As already mentioned, some of the operations used in Rijndael are based on \
the arithmetic of the finite field of 256 elements, GF(",
  Cell[BoxData[
      \(TraditionalForm\`2\^8\)]],
  "), which can be thought of as the set of all the binary polynomials (i.e., \
polynomials with coefficients in ",
  Cell[BoxData[
      \(TraditionalForm\`\[DoubleStruckCapitalZ]\_2\)]],
  "= GF(2), the field of two elements) of degree less than or equal to seven, \
with addition given by the ordinary polynomial sum (where the coefficients \
are added modulo 2) and multiplication induced by polynomial multiplication \
modulo the irreducible polynomial ",
  Cell[BoxData[
      \(TraditionalForm\`X\^8\)]],
  "+ ",
  Cell[BoxData[
      \(TraditionalForm\`X\^4\)]],
  "+ ",
  Cell[BoxData[
      \(TraditionalForm\`X\^3\)]],
  "+ ",
  Cell[BoxData[
      \(TraditionalForm\`X\)]],
  " + 1 (other irreducible polynomials could also have been used but this is \
the one that was chosen for Rijndael). Each polynomial of GF(",
  Cell[BoxData[
      \(TraditionalForm\`2\^8\)]],
  ") can be identified with the byte whose bits are the coefficients of the \
polynomial, with the least significant bit corresponding to the independent \
term and the most significant one corresponding to the leading coefficient. \
Thus, the underlying set of GF(",
  Cell[BoxData[
      \(TraditionalForm\`2\^8\)]],
  ") becomes identified with the set of all (unsigned) 8-bit integers {0, 1, \
2, ..., 255}, which we will refer to as the byte space. The field structure \
can be translated to this space, with the sum given by the bitwise XOR; the \
induced multiplication of bytes will be explicitly defined below. Thus, in \
the sequel, GF(",
  Cell[BoxData[
      \(TraditionalForm\`2\^8\)]],
  ") will be identified with the byte space, which is one of its many \
incarnations. We start by initializing  GF(",
  Cell[BoxData[
      \(TraditionalForm\`2\^8\)]],
  ") within the ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " package devoted to finite fields:"
}], "Text"],

Cell["\<\
<<Algebra`FiniteFields`;
ir = {1,1,0,1,1,0,0,0,1};
F = GF[2,ir];\
\>", "Input",
  InitializationCell->True],

Cell[TextData[{
  "Now, we construct functions that give the correspondence between bytes and \
polynomials in GF(",
  Cell[BoxData[
      \(TraditionalForm\`2\^8\)]],
  "). ",
  StyleBox["Mathematica's ",
    FontSlant->"Italic"],
  StyleBox["FiniteFields",
    FontFamily->"Courier New"],
  " package already contains built-in functions for these conversions but we \
define our own functions to do the job specifically for the field F. Since \
converting a list of bits to an element of F is much faster when the list has \
length 8 (i.e., includes all the polynomial coefficients, padded with zeros \
on the right) instead of some shorter length, we specify a length of 8 when \
calling ",
  StyleBox["IntegerDigits ",
    FontFamily->"Courier New"],
  StyleBox["from",
    FontFamily->"Times New Roman"],
  StyleBox[" ByteToFieldElement. ",
    FontFamily->"Courier New"],
  "Observe that, according to ",
  StyleBox["Mathematica'",
    FontSlant->"Italic"],
  "s convention, the list of coefficients of a polynomial starts with the \
independent term which, as already mentioned, corresponds to the least \
significant bit and this is the reason of using ",
  StyleBox["Reverse",
    FontFamily->"Courier New"],
  ". As these functions are only meant to be called from other functions \
which, hopefully, will supply them with data in the right format, we do not \
include any detailed conditional pattern matching here.  Both conversion \
functions are only used during the initialization stage."
}], "Text"],

Cell[BoxData[
    \(\(ByteToFieldElement[byte_Integer] := 
        F[Reverse[IntegerDigits[byte, 2, 8]]];\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(FieldElementToByte[element_] := 
        FromDigits[Reverse[element[\([1]\)]], 2];\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "Now we can define the multiplication in the byte space as follows (as \
already mentioned, the sum will be simply the bitwise XOR which, in ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  ", is given by the function ",
  StyleBox["BitXor",
    FontFamily->"Courier New"],
  "). We will not include this definition in the package, however, as we \
shall see that there are more efficient alternatives:\n\n\t",
  StyleBox["Mult[byte1_Integer,byte2_Integer]:=\n       \
Module[{el1=ByteToFieldElement[byte1], el2=ByteToFieldElement[byte2]},\n      \
\t If[byte1 byte2 \[Equal] 0, 0, FieldElementToByte[el1 el2]]];\n\t \
SetAttributes[Mult, Listable];",
    FontFamily->"Courier New"]
}], "Text"],

Cell[TextData[{
  "The operation ",
  StyleBox["MixColumns",
    FontFamily->"Courier New"],
  " of Rijndael makes use of matrix multiplication over GF(",
  Cell[BoxData[
      \(TraditionalForm\`2\^8\)]],
  ") and hence, in particular, of  multiplication in this field. However,  \
the multiplication function just defined is too slow for an intensive use \
like the one in Rijndael. We could improve the speed of this function by \
telling the finite fields package to do the field arithmetic (multiplication \
and computation of inverses) by representing the nonzero elements of the \
field as powers of a \"primitive element\" (more about this below). However, \
it would be inefficient to have to make the conversions from byte to field \
elements and back for each multiplication in Rijndael, so that it is \
preferable to implement the multiplication of bytes with a lookup table that \
will be generated at the beginning of the session; afterwards we will not \
deal with field elements anymore. We could simply use the multiplication just \
defined to generate this table but we can do it much more efficiently, \
without having to carry out all the multiplications. Indeed, suppose that the \
files of the table are labeled with the bytes in the 0-255 range and that we \
already know all the files up to file ",
  Cell[BoxData[
      \(TraditionalForm\`2\^i\)]],
  " for some 0 \[LessEqual] i \[LessEqual] 7. Then, since each element which \
is not a power of 2 is the sum of the previous power of 2 plus a smaller \
element, and  the multiplication is distributive with respect to the addition \
in GF(",
  Cell[BoxData[
      \(TraditionalForm\`2\^8\)]],
  "), we can compute the files ",
  Cell[BoxData[
      \(TraditionalForm\`2\^i\)]],
  "+1, ",
  Cell[BoxData[
      \(TraditionalForm\`2\^i\)]],
  "+2, ..., ",
  Cell[BoxData[
      \(TraditionalForm\`2\^\((i + 1)\)\)]],
  "-1 by xoring file ",
  Cell[BoxData[
      \(TraditionalForm\`2\^i\)]],
  " with files 1, 2, ..., ",
  Cell[BoxData[
      \(TraditionalForm\`2\^i\)]],
  "-1.  This way, we only need to use the multiplication of GF(",
  Cell[BoxData[
      \(TraditionalForm\`2\^8\)]],
  ") to compute the files corresponding to the powers of 2, namely, files 2,  \
",
  Cell[BoxData[
      \(TraditionalForm\`2\^2\)]],
  ", ..., ",
  Cell[BoxData[
      \(TraditionalForm\`2\^7\)]],
  "; the remaining ones are obtained with the much faster operation ",
  StyleBox["BitXor",
    FontFamily->"Courier New"],
  ".  Note that we cannot generate the rows of the table corresponding to \
powers of 2 by just xoring previous rows, because powers of 2 cannot be \
written as the sum of two smaller elements in GF(",
  Cell[BoxData[
      \(TraditionalForm\`2\^8\)]],
  "). The multiplication table of  GF(",
  Cell[BoxData[
      \(TraditionalForm\`2\^8\)]],
  ") has ",
  Cell[BoxData[
      \(TraditionalForm\`2\^8\)]],
  "\[Times] ",
  Cell[BoxData[
      \(TraditionalForm\`2\^8\)]],
  " integer entries and ",
  StyleBox["Mathematica'",
    FontSlant->"Italic"],
  "s function ",
  StyleBox["ByteCount",
    FontFamily->"Courier New"],
  " usually reports (on 32-bit computers, where four memory bytes are used \
for each machine integer) that it requires about 256KB of space storage, \
which is not much for today's computers. However, we do not need the full \
table, as the only bytes that appear in the matrices used by  ",
  StyleBox["MixColumns",
    FontFamily->"Courier New"],
  " and its inverse are 1, 2, 3, 9, 11, 13, 14. Then it would be enough to \
build up a reduced table consisting only of the files of the full table \
corresponding to these bytes. But this smaller table also has a disadvantage \
in that these scattered bytes require some not very attractive code which, on \
the other hand, would not be optimal from the point of view of speed. Thus we \
will settle for a somewhat larger but still compact multiplication table \
consisting of files 1-15 of the full table, which typically amounts to around \
16 KB of memory in ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  ". Summing up the preceding discussion, we have the following possibilities \
for generating this reduced table. The first one just uses the multiplication \
function defined above:\n\n\t",
  StyleBox["MultTable := Mult[#, Range[256]-1]&/@Range[15]\n\t\t",
    FontFamily->"Courier New"],
  "\n\tWhile this code, in the functional programming style of ",
  StyleBox["Mathematica,",
    FontSlant->"Italic"],
  " is certainly short, it is also rather slow and we can obtain the table \
much more efficiently by using ",
  StyleBox["BitXor",
    FontFamily->"Courier New"],
  " to generate, as explained above, the rows not corresponding to powers of \
2 (in imperative style this time):\n\t\n\t",
  StyleBox["MultTable := Module[{mt = Table[Range[256]-1,{15}]},\n\t\t\
Do[mt[[2^i]] = Mult[2^i,Range[256]-1],{i,1,3}];\n\t\tDo[mt[[j]] = \
BitXor[mt[[2^i]],mt[[j-2^i]]],{i,1,3},{j,2^i+1,2^(i+1)-1}];\n\t\t   mt];\t\n\t\
\t\t\t\t ",
    FontFamily->"Courier New"],
  "\n\tBut this code is not optimal either, as the multiplication by powers \
of 2 relies on going back and forth from bytes to field elements and using \
the multiplication of the ",
  StyleBox["FiniteFields",
    FontFamily->"Courier New"],
  " package, which is slow. In fact, the multiplication of bytes by powers of \
2 can be implemented much more efficiently, as it can be easily seen that \
each multiplication by 2 requires, at most, a bitwise (left) shift and a ",
  StyleBox["BitXor",
    FontFamily->"Courier New"],
  StyleBox[". The next function implements multiplication by 2 directly in \
the byte space, without relying on polynomial arithmetic;  the reason why ",
    FontFamily->"Times New Roman"],
  StyleBox["BitXor",
    FontFamily->"Courier New"],
  StyleBox[" with 283 is used comes from the fact that this is the value that \
the irreducible polynomial ",
    FontFamily->"Times New Roman"],
  Cell[BoxData[
      \(TraditionalForm\`X\^8\)]],
  "+ ",
  Cell[BoxData[
      \(TraditionalForm\`X\^4\)]],
  "+ ",
  Cell[BoxData[
      \(TraditionalForm\`X\^3\)]],
  "+ ",
  Cell[BoxData[
      \(TraditionalForm\`X\)]],
  " + 1, ",
  StyleBox["defining the field structure in the byte space, takes for ",
    FontFamily->"Times New Roman"],
  Cell[BoxData[
      \(TraditionalForm\`X\)]],
  StyleBox[" = 2:",
    FontFamily->"Times New Roman"]
}], "Text",
  TextAlignment->Left],

Cell[BoxData[{
    \(\(ByteTimes2[byte_Integer] := 
        If[byte < 128, 2\ byte, BitXor[283, 2\ byte]];\)\), "\n", 
    \(\(SetAttributes[ByteTimes2, Listable];\)\)}], "Input",
  InitializationCell->True],

Cell["\<\
Now, we can give our final version of the function generating the (reduced) \
multiplication table. It will be much faster than the method outlined above \
but, as the table is only generated once upon initialization, this is of no \
practical relevance in this situation. However, it shows how the \
multiplication table of a finite field of characteristic 2 can be generated \
by using only bit shifts and bit xors (with the obvious modifications it can \
generate the full table instead of the reduced one). In this case the field \
is very small and so the speed gain is irrelevant but with larger fields it \
could be important. But, of course, this would be useful only up to a certain \
point because, if the field is large, it may be completely impossible to \
generate the table.\
\>", "Text"],

Cell[BoxData[
    \(\(MultTable := 
        Module[{mt = Table[Range[256] - 1, {15}], 
            n = NestList[ByteTimes2, Range[256] - 1, 
                3]}, \[IndentingNewLine]Do[
            mt[\([2^i]\)] = n[\([i + 1]\)], {i, 1, 3}]; \[IndentingNewLine]Do[
            mt[\([j]\)] = BitXor[mt[\([2^i]\)], mt[\([j - 2^i]\)]], {i, 1, 
              3}, {j, 2^i + 1, 
              2^\((i + 1)\) - 
                1}]; \[IndentingNewLine]\ \ \ \ \ \ mt];\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "We would also like to point out an alternative to the preceding table for \
the multiplication in the byte space. Finite fields have the property that \
their multiplicative group (that is, the multiplicative group of nonzero \
elements of the field) is cyclic. This just means that all the nonzero \
elements can be obtained as powers of a primitive element (or a generator, in \
group-theoretic terms). In the case of the byte space, it can be easily seen \
that 3 is a primitive element (in fact, the smallest one) and that the list \
of the 255 nonzero bytes can be written in the form {",
  Cell[BoxData[
      \(TraditionalForm\`3\^0\)]],
  " = 1, ",
  Cell[BoxData[
      \(TraditionalForm\`3\^1\)]],
  " = 3, ",
  Cell[BoxData[
      \(TraditionalForm\`3\^2\)]],
  " =5, ",
  Cell[BoxData[
      \(TraditionalForm\`3\^3\)]],
  " = 15, ..., ",
  Cell[BoxData[
      \(TraditionalForm\`3\^254\)]],
  " = 246}. This list consists of the values of the \"discrete exponential \
function with base 3\" on the space of nonzero bytes and can be readily \
generated by using the function ",
  StyleBox["PowerList",
    FontFamily->"Courier New"],
  " defined in the finite fields package and converting the elements to bytes \
(or by using the function ",
  StyleBox["Mult",
    FontFamily->"Courier New"],
  " defined above, directly on the byte space). In fact, ",
  StyleBox["PowerList",
    FontFamily->"Courier New"],
  " automatically selects 3 as a primitive element in the case of our field \
F. The inverse of the discrete exponential function is the discrete logarithm \
(or index) function. The list of the (base 3) discrete logarithms of the \
nonzero bytes, {0, 25, 1, ..., 7},  can be obtained either by \"inverting\" \
the preceding list or by using ",
  StyleBox["Mathematica'",
    FontSlant->"Italic"],
  "s function ",
  StyleBox["FieldInd",
    FontFamily->"Courier New"],
  ". Once we have these two lists available, it is easy to implement the \
multiplication of  bytes as follows. To multiply bytes ",
  StyleBox["b",
    FontSlant->"Italic"],
  " and ",
  StyleBox["c",
    FontSlant->"Italic"],
  ", we fetch their logarithms (",
  StyleBox["x",
    FontSlant->"Italic"],
  " and ",
  StyleBox["y",
    FontSlant->"Italic"],
  ", say) from the latter list, so that ",
  StyleBox["b",
    FontSlant->"Italic"],
  " = ",
  Cell[BoxData[
      \(TraditionalForm\`3\^x\)]],
  " and ",
  StyleBox["c",
    FontSlant->"Italic"],
  " = ",
  Cell[BoxData[
      \(TraditionalForm\`3\^y\)]],
  ". Then, since ",
  StyleBox["bc",
    FontSlant->"Italic"],
  " = ",
  Cell[BoxData[
      \(TraditionalForm\`3\^\((x + y)\)\)]],
  ", where the exponent can be taken modulo 255,  all we have to do is to add \
",
  StyleBox["x",
    FontSlant->"Italic"],
  " plus ",
  StyleBox["y",
    FontSlant->"Italic"],
  " modulo 255 to obtain a result in the 0-254 range and look up in the first \
list to find the result. This method would perhaps be preferable in a \
space-constrained environment, as it only requires very little storage. \
However, the three lookups plus the addition would result in a somewhat \
slower multiplication function in our setting and this is the reason why we \
have opted for ",
  StyleBox["MultTable",
    FontFamily->"Courier New"],
  ". \n\tOn the other hand, as already mentioned in the introduction, the \
most efficient implementation of AES is based on 4 lookup tables (for all AES \
operations), each of which is a list of 256 4-byte words, as described in ",
  ButtonBox["[3]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  ", but we will not pursue this line in order to avoid excessive \
technicalities. \n\tSo far, we have managed to avoid the use of the ",
  StyleBox["FiniteFields",
    FontFamily->"Courier New"],
  " package but now we are going to need it for another function which is \
used in Rijndael, namely, the one that assigns to each nonzero element of \
GF(",
  Cell[BoxData[
      \(TraditionalForm\`2\^8\)]],
  ") its inverse (this could be programmed directly by using Euclid's \
algorithm for polynomials but the code would be a bit longer). We will use an \
extension of the inverse function which also takes 0 to 0 and, also in this \
case, we make it listable:"
}], "Text"],

Cell[BoxData[{
    \(\(InvByte[byte_Integer] := 
        If[byte \[Equal] 0, 0, 
          FieldElementToByte[
            1/ByteToFieldElement[byte]]];\)\), "\[IndentingNewLine]", 
    \(\(SetAttributes[InvByte, Listable];\)\)}], "Input",
  InitializationCell->True]
}, Closed]],

Cell[CellGroupData[{

Cell["Initialization of Rijndael constants and tables", "Section"],

Cell[TextData[{
  "Now we are going to initialize several objects which are needed in \
Rijndael and that we will store on static ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " variables as they will be generated only once per session and will be \
repeatedly used during the encryption and decryption stages. We start by \
constructing the \"",
  StyleBox["S Box",
    FontSlant->"Italic"],
  "\", which is needed for the ",
  StyleBox["SubBytes",
    FontFamily->"Courier New"],
  " operation to be defined later. The ",
  StyleBox["S Box",
    FontSlant->"Italic"],
  " is a permutation of the byte space which is defined by algebraic \
operations. These operations are the function that assigns to each (nonzero)  \
byte its inverse, given by the function ",
  StyleBox["InvByte",
    FontFamily->"Courier New"],
  " defined above, and an affine map consisting in multiplying each byte \
(regarded as a bit vector) by a matrix of bits and adding another 8-bit \
vector. We start by defining this vector, the matrix M, and its inverse IM."
}], "Text"],

Cell[BoxData[{
    \(\(vect = {1, 1, 0, 0, 0, 1, 1, 0};\)\), "\[IndentingNewLine]", 
    \(\(M = 
        Table[RotateRight[{1, 0, 0, 0, 1, 1, 1, 1}, i], {i, 0, 
            7}];\)\), "\[IndentingNewLine]", 
    \(\(IM = Inverse[M, Modulus \[Rule] 2];\)\)}], "Input",
  InitializationCell->True],

Cell[TextData[{
  "The affine map that will be applied after the InvByte function to define \
the ",
  StyleBox["SubBytes",
    FontFamily->"Courier New"],
  " operation is:"
}], "Text"],

Cell[BoxData[{
    \(\(AffMap[byte_Integer] := \ 
        Module[{coeffs = Reverse[IntegerDigits[byte, 2, 8]]}, 
          FromDigits[Reverse[Mod[M . coeffs + vect, 2]], 
            2]];\)\), "\[IndentingNewLine]", 
    \(\(SetAttributes[AffMap, \ Listable];\)\)}], "Input",
  InitializationCell->True],

Cell[TextData[{
  "Now, the permutation of the byte space to be used in the operation ",
  StyleBox["SubBytes",
    FontFamily->"Courier New"],
  " is given by the following function, which we call ",
  StyleBox["ByteSub",
    FontFamily->"Courier New"],
  ". We include it here to illustrate how ",
  StyleBox["SubBytes",
    FontFamily->"Courier New"],
  " works but we won't use this function as we will find a faster \
alternative; thus we leave it out of the package:\n\n\t",
  StyleBox["ByteSub[byte_Integer]:= AffMap[InvByte[byte]];\n \t  ",
    FontFamily->"Courier New"],
  "\nIt would be very inefficient to apply ",
  StyleBox["ByteSub",
    FontFamily->"Courier New"],
  " each time that a byte must be replaced by the ",
  StyleBox["SubBytes",
    FontFamily->"Courier New"],
  " operation. What we could do instead is to use this function to generate \
the ",
  StyleBox["S Box",
    FontSlant->"Italic"],
  " list and then use this list as a lookup table to implement the function \
",
  StyleBox["SubBytes",
    FontFamily->"Courier New"],
  " to be defined later. A short code to generate the ",
  StyleBox["S Box ",
    FontSlant->"Italic"],
  StyleBox["would then be:\n\n\t",
    FontFamily->"Times New Roman",
    FontVariations->{"CompatibilityType"->0}],
  StyleBox["S = ByteSub[#]& /@ Range[0, 255];",
    FontFamily->"Courier New"],
  "\n\nAgain, we left this out of the package because we can optimize the \
generation of the ",
  StyleBox["S Box ",
    FontSlant->"Italic"],
  StyleBox["by taking advantage of the simple structure of the permutation of \
the byte space defined by the function ",
    FontVariations->{"CompatibilityType"->0}],
  StyleBox["InvByte",
    FontFamily->"Courier New",
    FontVariations->{"CompatibilityType"->0}],
  StyleBox[". This permutation has order 2 (i.e., it gives the identity \
permutation when composed with itself) and hence it can be written as a \
product of disjoint transpositions. In other words, when you find the inverse \
of a byte, you actually find two inverses because you have also found the \
inverse of the inverse. This fact allows us to cut the number of computations \
of inverses in ",
    FontVariations->{"CompatibilityType"->0}],
  " GF(",
  Cell[BoxData[
      \(TraditionalForm\`2\^8\)]],
  ")",
  StyleBox[" needed to build the ",
    FontVariations->{"CompatibilityType"->0}],
  StyleBox["S Box",
    FontSlant->"Italic"],
  " in half (note that this remark does not apply to the permutation defined \
by the ",
  StyleBox["S Box",
    FontSlant->"Italic"],
  StyleBox[" itself, ",
    FontVariations->{"CompatibilityType"->0}],
  "which is just the ",
  StyleBox["SubBytes",
    FontFamily->"Courier New"],
  " operation of Rijndael; this permutation has a much more complicated cycle \
structure). Because of this fact, we split the computation of the ",
  StyleBox["S Box",
    FontSlant->"Italic"],
  " in two parts; first we compute the table of byte inverses with just 127 \
applications of the function ",
  StyleBox["InvByte",
    FontFamily->"Courier New"],
  ", and then we get the ",
  StyleBox["S Box",
    FontSlant->"Italic"],
  " by applying the affine map ",
  StyleBox["AffMap",
    FontFamily->"Courier New"],
  " to that table:"
}], "Text"],

Cell[BoxData[
    \(\(S = 
        Module[{x = \ Range[2, 255], \ s = Range[0, 255], a, b}, 
          Do[a = First[x]; b = InvByte[a]; s[\([a + 1]\)] = b; 
            s[\([b + 1]\)] = a; x = Complement[x, {a, b}], {127}]; 
          AffMap[s]];\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "Since the computation of inverses is the more time-consuming part, this \
approximately cuts in half the time required to compute the ",
  StyleBox["S Box",
    FontSlant->"Italic"],
  " by using ",
  StyleBox["ByteSub",
    FontFamily->"Courier New"],
  " as shown",
  StyleBox["; on today's computers this does not matter much as even the \
slower method outlined above is fast enough",
    FontFamily->"Times New Roman"],
  StyleBox[".",
    FontSlant->"Italic"],
  StyleBox[" Of course, an even faster method would be to just write down the \
",
    FontVariations->{"CompatibilityType"->0}],
  StyleBox["S Box",
    FontSlant->"Italic"],
  " in a list once and for all, instead of computing it each time. But we \
will refrain from doing so, if only for aesthetical and pedagogical reasons, \
and bearing also in mind that this computation must be carried out only once \
per session, during initialization. We will also need the inverse ",
  StyleBox["S Box",
    FontSlant->"Italic"],
  ", which is used in the decryption process. It can be obtained directly \
from the ",
  StyleBox["S Box",
    FontSlant->"Italic"],
  " and we will store it as \"IS\":"
}], "Text"],

Cell[BoxData[
    \(\(IS = 
        Module[{is = Range[256]}, 
          Do[is[\([i]\)] = \(Flatten[Position[S, i - 1]]\)[\([1]\)] - 1, {i, 
              256}]; is];\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "Next, we initialize the matrices used in ",
  StyleBox["MixColumns",
    FontFamily->"Courier New"],
  " and its inverse. The matrix of bytes to be used by ",
  StyleBox["MixColumns",
    FontFamily->"Courier New"],
  " is:"
}], "Text"],

Cell[BoxData[
    \(\(MC = Table[RotateRight[{2, 3, 1, 1}, i], {i, 0, 3}];\)\)], "Input",
  InitializationCell->True],

Cell["The inverse of MC is the matrix IMC, given by:", "Text"],

Cell[BoxData[
    \(\(IMC = Table[RotateRight[{14, 11, 13, 9}, i], {i, 0, 3}];\)\)], "Input",\

  InitializationCell->True],

Cell[TextData[{
  "For the operation ",
  StyleBox["MixColumns",
    FontFamily->"Courier New"],
  " we will also need the function that multiplies bytes based on the \
multiplication table of the byte space. We initialize the table:"
}], "Text"],

Cell[BoxData[
    \(\(MT = MultTable;\)\)], "Input",
  InitializationCell->True],

Cell["\<\
Now, we define the \"fast multiplication of bytes\" based on the previous \
\"reduced multiplication table\", where the first row starts with byte 1 \
instead of byte 0. We make this function listable:\
\>", "Text"],

Cell[BoxData[{
    \(\(ByteMult[byte1_Integer, 
          byte2_Integer] := \(MT[\([byte1]\)]\)[\([byte2 + 
              1]\)];\)\), "\[IndentingNewLine]", 
    \(\(SetAttributes[ByteMult, Listable];\)\)}], "Input",
  InitializationCell->True],

Cell[TextData[{
  "For the ",
  StyleBox["KeyExpansion",
    FontFamily->"Courier New"],
  " operation, the so-called \"round constants\" are needed (cf. ",
  ButtonBox["[3]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  "). They are given by:"
}], "Text"],

Cell[BoxData[
    \(\(Rcon = NestList[ByteMult[2, #] &, {1, 0, 0, 0}, 13];\)\)], "Input",
  InitializationCell->True]
}, Closed]],

Cell[CellGroupData[{

Cell["Rijndael Operations", "Section"],

Cell[TextData[{
  "A block of plaintext to be encrypted by Rijndael consists of 128 bits \
which can be decomposed into 16 bytes. These 16 bytes are arranged columnwise \
in a 4x4 matrix called the state. The operations of Rijndael act on state in \
different ways. For example, ",
  StyleBox["SubBytes",
    FontFamily->"Courier New"],
  " applies a permutation (or substitution) of the byte space to each byte \
independently, so that it transforms each byte always in the same way. This \
permutation is given by the ",
  StyleBox["S Box",
    FontSlant->"Italic"],
  " which has been already defined. Other operations act more globally on the \
state matrix, as it happens with ",
  StyleBox["ShiftRows",
    FontFamily->"Courier New"],
  " and ",
  StyleBox["MixColumns",
    FontFamily->"Courier New"],
  ". ",
  StyleBox["AddRoundKey",
    FontFamily->"Courier New"],
  ", on the other hand, performs a ",
  StyleBox["BitXor",
    FontFamily->"Courier New"],
  " of the bytes of state with those of a matrix of the same size obtained by \
means of the ",
  StyleBox["KeyExpansion",
    FontFamily->"Courier New"],
  " process. In this case, the transformation applied to each byte depends on \
several things: the key, the position of the byte in state, and the round \
being performed. We define the corresponding functions in such a way that \
they can act on the list of all states (i.e., on the complete plaintext) at \
once. In the case of ",
  StyleBox["SubBytes",
    FontFamily->"Courier New"],
  " and its inverse this is achieved simply by making the corresponding \
functions listable. In the case of operations that act more globally on \
state, we use a \"pure function\" to allow these operations to act on the \
list of all states."
}], "Text"],

Cell[CellGroupData[{

Cell["SubBytes and its inverse", "Subsection"],

Cell[TextData[{
  "We now define the ",
  StyleBox["SubBytes",
    FontFamily->"Courier New"],
  " operation by means of a function with the same name. It defines the same \
permutation of the byte space as the one given by the function ",
  StyleBox["ByteSub",
    FontFamily->"Courier New"],
  " above, but ",
  StyleBox["SubBytes",
    FontFamily->"Courier New"],
  " is much faster because it just looks in the ",
  StyleBox["S Box",
    FontSlant->"Italic"],
  " for the value corresponding to each byte."
}], "Text"],

Cell[BoxData[{
    \(\(SubBytes[byte_Integer] := 
        S[\([byte + 1]\)];\)\), "\[IndentingNewLine]", 
    \(SetAttributes[SubBytes, Listable]\)}], "Input",
  InitializationCell->True],

Cell[TextData[{
  "The inverse operation just looks in the inverse ",
  StyleBox["S Box",
    FontSlant->"Italic"],
  ":"
}], "Text"],

Cell[BoxData[{
    \(\(InvSubBytes[byte_Integer] := 
        IS[\([byte + 1]\)];\)\), "\[IndentingNewLine]", 
    \(SetAttributes[InvSubBytes, Listable]\)}], "Input",
  InitializationCell->True]
}, Closed]],

Cell[CellGroupData[{

Cell["ShiftRows and its inverse", "Subsection"],

Cell[TextData[{
  "The operation ",
  StyleBox["ShiftRows",
    FontFamily->"Courier New"],
  " acts on the rows of the state matrix. Since a matrix is a list, making ",
  StyleBox["ShiftRows",
    FontFamily->"Courier New"],
  " listable will not work. However, we would like it to be listable on the \
first level of lists, in order to be able to apply the operation to all \
states at once. For this we set:"
}], "Text"],

Cell[BoxData[
    \(\(ShiftLeft[mat_List] := 
        Module[{m = mat}, 
          Do[m[\([i]\)] = RotateLeft[m[\([i]\)], i - 1], {i, 2, 4}]; 
          m];\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(ShiftRows[states_List] := \(ShiftLeft[#] &\) /@ states\)], "Input",
  InitializationCell->True],

Cell["\<\
The inverse is similar, only with right shifting instead of left shifting:\
\>", "Text"],

Cell[BoxData[
    \(\(ShiftRight[mat_List] := 
        Module[{m = mat}, 
          Do[m[\([i]\)] = RotateRight[m[\([i]\)], i - 1], {i, 2, 4}]; 
          m];\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(InvShiftRows[states_List] := \(ShiftRight[#] &\) /@ states\)], "Input",
  InitializationCell->True]
}, Closed]],

Cell[CellGroupData[{

Cell["MixColumns and its inverse", "Subsection"],

Cell[TextData[{
  " ",
  StyleBox["MixColumns",
    FontFamily->"Courier New"],
  " and its inverse are really short based on ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  "'s powerful ",
  StyleBox["Inner",
    FontFamily->"Courier New"],
  " function which makes easy to define the multiplication of matrices whose \
entries are in GF(",
  Cell[BoxData[
      \(TraditionalForm\`2\^8\)]],
  ") -regarded as the byte space- by using the operations of this field. For \
each state (i.e., for each plaintext block), ",
  StyleBox["MixColumns",
    FontFamily->"Courier New"],
  " requires 64 multiplications of bytes per round, i.e., a minimum of  576 \
multiplications. This is the reason why using ",
  StyleBox["ByteMult",
    FontFamily->"Courier New"],
  " instead of ",
  StyleBox["Mult",
    FontFamily->"Courier New"],
  " for the multiplication gives a huge improvement in the overall speed of \
the Rijndael encryption (and decryption) algorithm."
}], "Text"],

Cell[BoxData[
    \(MixColumns[states_List] := \(Inner[ByteMult, MC, #, BitXor] &\) /@ 
        states\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(InvMixColumns[states_List] := \(Inner[ByteMult, IMC, #, BitXor] &\) /@ 
        states\)], "Input",
  InitializationCell->True]
}, Closed]],

Cell[CellGroupData[{

Cell["Generation of the round keys", "Subsection"],

Cell[TextData[{
  "Now we set up the Key schedule process. The input is the Rijndael key as a \
list of 16, 24 or 32 bytes. The output is a list of 11, 13 or 15 matrices of \
size 4x4 whose coefficients are bytes. The first 4, 6 or 8 columns of the \
first two matrices contain the key, which has been mapped to these matrices \
columnwise, starting on the top of the leftmost column.  The functions ",
  StyleBox["RotWord",
    FontFamily->"Courier New"],
  " and ",
  StyleBox["SubWord",
    FontFamily->"Courier New"],
  " mentioned in the description of ",
  StyleBox["KeyExpansion",
    FontFamily->"Courier New"],
  " given in ",
  ButtonBox["[11]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  " are not defined here because ",
  StyleBox["RotWord",
    FontFamily->"Courier New"],
  " is simply ",
  StyleBox["Mathematica'",
    FontSlant->"Italic"],
  "s function ",
  StyleBox["RotateLeft",
    FontFamily->"Courier New"],
  " and ",
  StyleBox["SubWord",
    FontFamily->"Courier New"],
  " is ",
  StyleBox["SubBytes",
    FontFamily->"Courier New"],
  " applied to a word of 4 bytes (which is done here automatically, as ",
  StyleBox["SubBytes",
    FontFamily->"Courier New"],
  " has the attribute \"Listable\"). The following is an almost verbatim \
transcription of the pseudocode in ",
  ButtonBox["[3]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  ":"
}], "Text"],

Cell[BoxData[
    \(\(KeyExpansion[
          key_List] := \[IndentingNewLine]Module[{l = Length[key], 
            k = Partition[key, 4], s, w, temp}, \[IndentingNewLine]s = 
            l + 28; w = Table[{0, 0, 0, 0}, {s}]; 
          Nk = l/4; \[IndentingNewLine]Do[w[\([i]\)] = k[\([i]\)], {i, Nk}]; 
          Do[temp = w[\([i - 1]\)]; 
            If[\ Mod[i - 1, Nk] \[Equal] 0, 
              temp = BitXor[SubBytes[RotateLeft[temp]], 
                  Rcon[\([\((i - 1)\)/Nk]\)]], 
              If[Nk \[Equal] 8 && Mod[i - 5, Nk] \[Equal] 0, 
                temp = SubBytes[temp]]]; 
            w[\([i]\)] = BitXor[w[\([i - Nk]\)], temp], {i, Nk + 1, 
              s}]; \[IndentingNewLine]\(Transpose[#] &\) /@ 
            Partition[w, 4]];\)\)], "Input",
  InitializationCell->True]
}, Closed]],

Cell[CellGroupData[{

Cell["AddRoundKey", "Subsection"],

Cell[TextData[{
  "The operation ",
  StyleBox["AddRoundKey",
    FontFamily->"Courier New"],
  " is simply the ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " function ",
  StyleBox["BitXor",
    FontFamily->"Courier New"],
  ". We define it in a way that allows it to act on the list of all states:"
}], "Text"],

Cell[BoxData[
    \(AddRoundKey[key_List, states_List] := \(BitXor[key, #] &\) /@ 
        states\)], "Input",
  InitializationCell->True]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Encryption and Decryption functions", "Section"],

Cell["\<\
The Rijndael operations previously defined are ready to act on a list of \
states, so that the encryption and decryption functions of AES can be \
obtained from them in a straightforward way. These functions accept as input \
the message to be encrypted, given as a list of bytes whose length must be a \
multiple of 16, and the cipher key, also as a list of bytes of length 16, 24, \
or 32. Later we will build functions that convert the message strings or \
files to be encrypted to this format and conversely. First we give functions \
that convert a list of bytes to a list of states and back.\
\>", "Text"],

Cell[BoxData[
    \(\(BytesToStates[bytes_List] := \(Transpose[Partition[#, 4]] &\) /@ 
          Partition[bytes, 16];\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(StatesToBytes[states_List] := 
        Flatten[\(Transpose[#] &\) /@ states];\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "Next we define an encryption round of Rijndael. This will be used for \
rounds 1, ..., Nr-1 (where Nr is the total number of rounds). The last round \
is different because the ",
  StyleBox["MixColumns",
    FontFamily->"Courier New"],
  " operation is omitted."
}], "Text"],

Cell[BoxData[
    \(\(RijndaelRound[states_List, roundk_List] := 
        AddRoundKey[roundk, 
          MixColumns[ShiftRows[SubBytes[states]]]];\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "We are now ready to define the encryption function used by Rijndael. This \
is often referred to as the ",
  StyleBox["forward cipher function",
    FontSlant->"Italic"],
  " (cf. ",
  ButtonBox["[5]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  ") to distinguish it from the higher level encryption functions that will \
be used to encrypt text strings or files. The input is the plaintext and the \
key, both of them given as a list of bytes. The plaintext is converted to a \
list of states which are encrypted and converted to a list of bytes that \
makes up the output ciphertext."
}], "Text"],

Cell[BoxData[
    \(\(RijndaelEncrypt[bytes_List, key_List] := 
        Module[{ek = KeyExpansion[key], s = BytesToStates[bytes], 
            Nr = Length[key]/4 + 6}, \[IndentingNewLine]s = 
            AddRoundKey[ek[\([1]\)], s]; \[IndentingNewLine]s = 
            Fold[RijndaelRound[#1, #2] &, s, 
              Drop[Rest[ek], \(-1\)]]; \[IndentingNewLine]s = 
            AddRoundKey[ek[\([Nr + 1]\)], 
              ShiftRows[
                SubBytes[s]]]; \[IndentingNewLine]\ \ \ \ \ \ \ StatesToBytes[
            s]];\)\)], "Input",
  InitializationCell->True],

Cell["\<\
For the decryption function (or inverse cipher function), the inverse \
operations are used in reverse order:\
\>", "Text"],

Cell[BoxData[
    \(\(RijndaelInvRound[states_List, roundk_List] := 
        InvSubBytes[
          InvShiftRows[
            InvMixColumns[AddRoundKey[roundk, states]]]];\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(RijndaelDecrypt[bytes_List, key_List] := 
        Module[{ek = KeyExpansion[key], s = BytesToStates[bytes], 
            Nr = Length[key]/4 + 6}, \[IndentingNewLine]s = 
            InvSubBytes[
              InvShiftRows[
                AddRoundKey[ek[\([Nr + 1]\)], s]]]; \[IndentingNewLine]s = 
            Fold[RijndaelInvRound[#1, #2] &, s, 
              Reverse[Drop[Rest[ek], \(-1\)]]]; \[IndentingNewLine]s = 
            AddRoundKey[ek[\([1]\)], 
              s]; \[IndentingNewLine]\ \ \ \ \ \ \ StatesToBytes[
            s]];\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "The encryption and decryption functions previously defined are all we need \
for the ECB mode of AES. However, for CBC and CFB modes we have to define new \
encryption and decryption functions that will act on top of them and this is \
very easy with ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  "'s functions ",
  StyleBox["FoldList",
    FontFamily->"Courier New"],
  " and ",
  StyleBox["MapThread",
    FontFamily->"Courier New"],
  ". The encryption functions make use of an initialization vector (IV) that \
must be unpredictable ",
  ButtonBox["[5]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  " but for now we just will pass the IV as an argument to these functions, \
without caring about its nature. This way, the encryption functions will \
later be used either with random IV's (for actual encryption) or with \
specified IV's (for checking purposes). The ciphertext produced by the \
encryption functions is, as before, a list of bytes but this time it contains \
the IV in the first 16 positions. First, the encryption function for CBC \
mode:"
}], "Text"],

Cell[BoxData[
    \(\(CbcEncrypt[bytes_List, key_List, iv_List] := 
        Module[{b = Partition[Join[iv, bytes], 16], 
            k = key}, \[IndentingNewLine]Flatten[
            FoldList[RijndaelEncrypt[BitXor[#1, #2], k] &, First[b], 
              Rest[b]]]];\)\)], "Input",
  InitializationCell->True],

Cell["The decryption function for CBC mode is then the following:", "Text"],

Cell[BoxData[
    \(\(CbcDecrypt[bytes_List, 
          key_List] := \[IndentingNewLine]Module[{b = Partition[bytes, 16], 
            k = key}, 
          Flatten[MapThread[
              BitXor[#1, RijndaelDecrypt[#2, k]] &, {Drop[b, \(-1\)], 
                Rest[b]}]]];\)\)], "Input",
  InitializationCell->True],

Cell["\<\
The CFB encryption and decryption functions are very similar because we are \
using the 128-bit CFB mode and so these functions deal with the same \
plaintext blocks as in the ECB mode:\
\>", "Text"],

Cell[BoxData[
    \(\(CfbEncrypt[bytes_List, key_List, \ iv_List] := 
        Module[{b = Partition[Join[iv, bytes], 16], 
            k = key}, \[IndentingNewLine]Flatten[
            FoldList[BitXor[RijndaelEncrypt[#1, k], #2] &, First[b], 
              Rest[b]]]];\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(CfbDecrypt[bytes_List, 
          key_List] := \[IndentingNewLine]Module[{b = Partition[bytes, 16], 
            k = key}, 
          Flatten[MapThread[
              BitXor[RijndaelEncrypt[#1, k], #2] &, {Drop[b, \(-1\)], 
                Rest[b]}]]];\)\)], "Input",
  InitializationCell->True],

Cell["\<\
The  preceding encryption and decryption functions for the different modes, \
can  be now combined in order to allow the selection of the encryption or \
decryption mode:\
\>", "Text"],

Cell[BoxData[
    \(\(RijndaelEncryptMode[bytes_List, key_List, mode_String, iv_List] := 
        Module[{b = bytes, k = key, m = mode, i = iv}, 
          Which[m == "\<cbc\>", 
            CbcEncrypt[b, k, 
              i], \[IndentingNewLine]\ \ \ \ \ \ \ \ \ \ \ \ m == "\<ecb\>", 
            RijndaelEncrypt[b, 
              k], \ \ \ \ \ \ \ \ \ \ \ \ \[IndentingNewLine]\ \ \ \ \ \ \ \ \
\ \ \ \ True, \ CfbEncrypt[b, k, i]]];\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(RijndaelDecryptMode[bytes_List, key_List, mode_String] := 
        Module[{b = bytes, k = key, m = mode}, 
          Which[m == "\<cbc\>", 
            CbcDecrypt[b, 
              k], \[IndentingNewLine]\ \ \ \ \ \ \ \ \ \ \ \ m == "\<ecb\>", 
            RijndaelDecrypt[b, 
              k], \ \ \ \ \ \ \ \ \ \ \ \ \[IndentingNewLine]\ \ \ \ \ \ \ \ \
\ \ \ \ True, \ CfbDecrypt[b, k]]];\)\)], "Input",
  InitializationCell->True]
}, Closed]],

Cell[CellGroupData[{

Cell["Input and output functions", "Section"],

Cell["\<\
First of all, we define functions to pad a list of bytes to be used as input \
by the encryption function and to do the inverse operation to recover the \
original list after applying the decryption function. The message is padded \
on the right with random bytes to make the total byte length a multiple of \
16. To keep track of the true length of the message, we encode the number of \
extra bytes in the last byte added. To avoid ambiguity when the length of the \
original message is already a multiple of 16, a full 16-byte block is added \
in this case.\
\>", "Text"],

Cell[BoxData[
    \(\(PadByteList[bytes_?\((VectorQ[#, IntegerQ] &)\)] := 
        Module[{b = bytes, l, rem}, \[IndentingNewLine]l = Length[b]; 
          rem = Mod[l, 16]; \ 
          PadRight[b, l + 16 - rem, 
            Join[Table[
                Random[Integer, 255], {15}], {16 - rem}]]];\)\)], "Input",
  InitializationCell->True],

Cell["\<\
The next function recovers the original list of bytes by removing the \
padding. Since this function will only be applied to a list of bytes obtained \
from a Rijndael decryption, the last byte should be a positive integer less \
than or equal to 16. We check whether this is indeed the case for, otherwise, \
it means that an error has occurred, probably due to having used a wrong \
decryption key or having decrypted in the wrong mode. This check will permit \
to add an error message to the decryption functions and stop them gracefully \
is this error is detected.\
\>", "Text"],

Cell[BoxData[
    \(\(UnpadByteList[bytes_?\((VectorQ[#, IntegerQ] &)\)] := 
        Module[{b = \ bytes, l = Last[bytes]}, 
          If\ [l > 16 || l \[Equal] 0, Return[], b = Drop[b, \(-l\)]]; 
          b];\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "Observe that padding with random bytes has, in the case of very  short \
messages of less than 15 characters, a beneficial side effect in relation \
with the main weakness of ECB mode, namely, the fact that identical \
plaintexts encrypted with the same key yield identical ciphertexts. If we \
have a text string (or a file) of less than 15 characters, we will convert it \
to a list of bytes which will then be padded. The list obtained after padding \
will be different most of the time,  this being more likely the shorter the \
message is. Accordingly, the ciphertexts obtained from such a text string in \
different encryptions with the same key, will also be different in general. \
Of course, regardless of plaintext size, the ciphertext blocks will almost \
always be different if CBC or CFB mode is used.\n\tWe will provide functions \
to encrypt and decrypt not only text strings but also binary files, so that \
we need a function to read from disk a binary file to be encrypted and \
another one to write the encrypted file to disk afterwards. To read binary \
files from disk, ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " has, from version 5.1, the function ",
  StyleBox["BinaryReadList,",
    FontFamily->"Courier New"],
  " which is very convenient for our purposes since it will read a file and \
produce a list of its bytes. However,  in versions prior to 5.1 this function \
is not available and we use ",
  StyleBox["BinaryImport",
    FontFamily->"Courier New"],
  ", from the \"",
  StyleBox["Experimental`",
    FontFamily->"Courier New"],
  "\"  context. The next function reads a file and converts it to a list of \
bytes:"
}], "Text"],

Cell[BoxData[
    \(\(FileToByteList[filename_String] := 
        Module[{b}, 
          If[$VersionNumber >= \ 5.1\ , \ b = BinaryReadList[filename], 
            b = Experimental`BinaryImport[filename, {"\<Byte\>" .. }]; 
            If[\ b \[Equal] {EndOfFile}, b = {}]]; b];\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "The inverse of ",
  StyleBox["FileToByteList",
    FontFamily->"Courier New"],
  " is the next function which also checks the ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " version to choose which method to use. For ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " versions prior to 5.1, ",
  StyleBox["OpenWrite",
    FontFamily->"Courier New"],
  " and ",
  StyleBox["Close",
    FontFamily->"Courier New"],
  " are used to obtain a 0-byte file when writing to disk the empty list. ",
  StyleBox["BinaryExport",
    FontFamily->"Courier New"],
  " does it too but prints an error message that could be confusing to the \
user. A filename must be supplied to store the file on disk."
}], "Text"],

Cell[BoxData[
    \(\(ByteListToFile[bytes_?\((VectorQ[#, IntegerQ] &)\), 
          filename_String] := \[IndentingNewLine]Module[{b = bytes}, 
          If[$VersionNumber >= \ 5.1, BinaryWrite[filename, b, "\<Byte\>"]; 
            Close[filename], \[IndentingNewLine]If[b \[Equal] {}, 
              OpenWrite[filename]; 
              Close[filename], \[IndentingNewLine]Experimental`BinaryExport[
                filename, b, {"\<Byte\>" .. }]]]];\)\)], "Input",
  InitializationCell->True],

Cell["\<\
The next function allows one to generate a Rijndael key from a nonempty text \
string. If the text string has 16 bytes or more, the first 16 bytes are \
taken, otherwise, the string is appended to itself as many times as \
necessary. The required length of the key must be supplied as the second \
argument to the function, which will only accept one of the three allowed \
keylengths.\
\>", "Text"],

Cell[BoxData[
    \(\(TextToKey[text_String, length_] := 
        Module[{t = text, l = length}, 
          If[StringLength[t] \[Equal] 0, Message[TextToKey::noemp]; 
            Return[]]; 
          If[MemberQ[{128, 192, 256}, l], \[IndentingNewLine]While[
              StringLength[t] < l/8, t = StringJoin[t, t]]; 
            ToCharacterCode[StringTake[t, l/8]], Message[TextToKey::noleng]; 
            Return[]]];\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "The preceding method of generating a key is not very recommendable and is \
included here only for historical reasons. First, if the characters are \
supplied through the keyboard by typing in the standard way, the number of \
different key bytes that can be obtained is severely restricted. Moreover,  \
if one uses a meaningful sentence in a natural language, then the effective \
size of the key space is further reduced -and even more so if a very short \
text string is used. This would violate a basic cryptographic principle which \
postulates that all keys should be equally likely and they should be chosen \
uniformly at random ",
  ButtonBox["[7]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  ". Thus, in general, it would be preferable to use as key a (pseudo-)random \
list of  bytes that can be generated by:"
}], "Text"],

Cell[BoxData[
    \(\(RandomKey[length_] := 
        If[MemberQ[{128, 192, 256}, length], 
          Table[Random[Integer, 255], {length/8}], \[IndentingNewLine]Message[
            RandomKey::noleng]; Return[]];\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "In the encryption and decryption functions to be defined below, the key \
will be often supplied as a list of bytes, preferably generated by ",
  StyleBox["RandomKey",
    FontFamily->"Courier New"],
  StyleBox[". We will check whether the key is valid by checking that it is \
indeed a list of bytes of one of the allowed lengths:",
    FontFamily->"Times New Roman"]
}], "Text"],

Cell[BoxData[
    \(\(CheckKey[key_List] := 
        Module[{k = key}, 
          If[\(! VectorQ[k, \((0 \[LessEqual] \ # \[LessEqual] \ 255)\) &]\), 
            Message[CheckKey::nobytes, key]; 
            Return[]]; \[IndentingNewLine]If[\(! MemberQ[{16, 24, 32}, 
                Length[k]]\), Message[CheckKey::noleng, key]; Return[]]; 
          k];\)\)], "Input",
  InitializationCell->True],

Cell["\<\
The next function will be used to check that the mode supplied to an \
encryption or decryption function is one of the three allowed ones and has \
not been mistyped. If the selected mode is not valid, the user is prompted \
for a valid one or the operation can be aborted.\
\>", "Text"],

Cell[BoxData[
    \(\(CheckMode[mode_String] := 
        Module[{prompt, m = mode}, 
          While[m \[NotEqual] "\<cbc\>" && m \[NotEqual] "\<ecb\>" && 
              m \[NotEqual] "\<cfb\>", \[IndentingNewLine]prompt = 
              StringJoin["\<Mode \>", m, "\< is not valid. Please input: \r\n
      cbc  for CBC mode \r\n
      ecb  for ECB mode \r\n
      cfb  for CFB mode \r\n
      a  to abort\>"]; \ \n\ \ \ \ \ \ \ \ \ m = 
              InputString[prompt]; \[IndentingNewLine]\ \ If[
              m \[Equal] "\<a\>", Return[]]]; m];\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "Now, the encryption and decryption AES functions, both for text and binary \
files. Since Rijndael can be viewed as the algorithm underlying AES, we will \
adopt the convention of using the term \"AES\" in the name of the higher \
level functions that are exported from the package and accessible to the user \
once the package is loaded, whereas \"Rijndael\" will only be used in the \
name of non-exported functions (with the sole exception of the function ",
  StyleBox["DisplayRijndaelObject",
    FontFamily->"Courier New"],
  " defined below). When dealing with text, both the plaintext and the \
encrypted message are text strings. The plaintext can be typed directly from \
the keyboard or it can be read from disk using the ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " function ",
  StyleBox["Read",
    FontFamily->"Courier New"],
  ". Then it is converted to a list of bytes with the help of the ISO Latin-1 \
code and padded on the right. The ciphertext can be displayed on screen \
and/or written to disk using, for example, ",
  StyleBox["Mathematica'",
    FontSlant->"Italic"],
  "s function ",
  StyleBox["Write",
    FontFamily->"Courier New"],
  ". The  software automatically detects the size of the key, which should \
have been generated by ",
  StyleBox["TextToKey",
    FontFamily->"Courier New"],
  " or ",
  StyleBox["RandomKey",
    FontFamily->"Courier New"],
  " and not typed directly on the keyboard to minimize the possibility of \
errors. Moreover, these functions will provide only keys with the correct \
sizes and in the required format.\n\tWe start with the encryption function \
for text strings. It is perhaps worthwhile to mention that whenever we speak \
of  \"text strings\", we mean strings of characters whose codes belong to the \
0-255 range and hence they can include \"non-printable\" characters. Apart \
from the plaintext and the key, these functions accept an optional third \
argument for the mode (which should be supplied in lower case letters as \
\"ecb\", \"cbc\" or \"cfb\"). If this argument is left blank, CBC mode is \
used by default. In general, CBC mode is preferable to ECB mode in most \
situations and moreover, it can be used for authentication purposes. Thus we \
will use CBC mode as default in all the encryption and decryption functions \
where the mode can be selected. On the other hand, we use ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  "'s pseudo-random number generator to generate the unpredictable IV \
required by CBC and CFB modes just before encryption.\t"
}], "Text"],

Cell[BoxData[
    \(\(AesTextEncrypt[plaintext_String, key_List, 
          mode_String: "\<cbc\>"] := \[IndentingNewLine]Module[{b = 
              ToCharacterCode[plaintext], k = CheckKey[key], 
            m = CheckMode[mode], 
            iv = Table[Random[Integer, 255], {16}]}, \[IndentingNewLine]If[
            k === Null, Return[]]; \[IndentingNewLine]If[m\  === \ Null, 
            Print["\<Encryption aborted by user\>"]; 
            Return[]]; \[IndentingNewLine]FromCharacterCode[
            RijndaelEncryptMode[PadByteList[b], k, m, iv]]];\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "In the decryption function we check the last byte of the decrypted message \
through the output of the ",
  StyleBox["BytesToText",
    FontFamily->"Courier New"],
  " function. If this byte is outside the expected range, an error message is \
printed out and the operation is stopped. This action is also taken if the \
byte length of the ciphertext string is not a multiple of 16."
}], "Text"],

Cell[BoxData[
    \(\(AesTextDecrypt[ciphertext_String, key_List, mode_String: "\<cbc\>"] := 
        Module[{c = ToCharacterCode[ciphertext], k = CheckKey[key], 
            m = CheckMode[mode]}, \[IndentingNewLine]If[k === Null, 
            Return[]]; \[IndentingNewLine]If[m\  === \ Null, 
            Print["\<Decryption aborted by user\>"]; 
            Return[]]; \[IndentingNewLine]If[\ 
            Mod[Length[c], 16] \[NotEqual] 0, 
            Message[AesTextDecrypt::notaes, ciphertext]; 
            Return[]]; \[IndentingNewLine]c = 
            UnpadByteList[
              RijndaelDecryptMode[c, k, m]]; \[IndentingNewLine]If[
            c === Null, Message[AesTextDecrypt::badbyte]; Return[], 
            FromCharacterCode[c]]];\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "In the binary case, the functions to be used are similar but, before \
writing a file to disk, we will check if a file with the same name already \
exists. This will be done with the next function which, in this case, will \
prompt the user asking whether the file should be overwritten or a different \
name should be used, giving also the option of aborting the operation. The \
output of this function is either ",
  StyleBox["Null",
    FontFamily->"Courier New"],
  " (if the user chooses to abort) or the file name to be used when writing \
to disk."
}], "Text"],

Cell[BoxData[
    \(\(CheckFileName[
          filename_String] := \[IndentingNewLine]Module[{answer, prompt, 
            f = filename}, \[IndentingNewLine]While[
            FileNames[f] \[NotEqual] {}, \[IndentingNewLine]prompt = 
              StringJoin["\<A file named \>", 
                f, "\< already exists. Please input: \r\n 
o  to overwrite \r\n 
c  to choose another file name \r\n 
anything else to abort\>"]; \[IndentingNewLine]answer = 
              InputString[
                prompt]; \[IndentingNewLine]Which[\[IndentingNewLine]\ 
              answer == "\<o\>", Break[], \[IndentingNewLine]\ 
              answer \[Equal] "\<c\>", prompt = "\<Enter new file name:\>"; 
              f = InputString[prompt], \[IndentingNewLine]\ True, Return[]]]; 
          f];\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "Next, the encryption and decryption functions for binary files; for \
inputting the keys, the previous remarks for the case of the text encrypting \
and decrypting functions are also applicable here. Like the encryption and \
decryption functions for text strings, these functions must be supplied with \
a file name, a key and, optionally, the mode to be used (if not specified, \
CBC mode is used by default). Moreover, there is an option named ",
  StyleBox["CheckFile",
    FontFamily->"Courier New"],
  " that uses ",
  StyleBox["CheckFileName",
    FontFamily->"Courier New"],
  " by default to check whether a file with the supplied name exists. In that \
case, the user can choose between aborting the operation, overwriting the \
file, or using a different file name. On the other hand, if the option ",
  StyleBox["CheckFile",
    FontFamily->"Courier New"],
  " is set to ",
  StyleBox["False",
    FontFamily->"Courier New"],
  ", then the file will be written to disk under the supplied name, without \
checking whether a file with the same name already exists. This is useful to \
do some checking as in one of the functions to be defined below."
}], "Text"],

Cell[BoxData[
    \(\(AesBinaryEncrypt[filename_String, cfilename_String, key_List, 
          mode_String: "\<cbc\>", \ 
          opts___Rule] := \[IndentingNewLine]Module[{b = 
              FileToByteList[filename], cf = cfilename, k = CheckKey[key], 
            m = CheckMode[mode], iv = Table[Random[Integer, 255], {16}], 
            ch = \(CheckFile /. {opts}\) /. \ 
                Options[AesBinaryEncrypt]}, \[IndentingNewLine]If[
            k === Null || b === $Failed, \ Return[]]; \[IndentingNewLine]If[
            m === Null, \ Print["\<Encryption aborted by user\>"]; 
            Return[]]; \[IndentingNewLine]If[ch, cf = CheckFileName[cf]; 
            If[cf === Null, Print["\<Encryption aborted by user\>"]; 
              Return[]]]; \[IndentingNewLine]ByteListToFile[
            RijndaelEncryptMode[PadByteList[b], k, m, iv], cf]];\)\)], "Input",\

  InitializationCell->True],

Cell[BoxData[
    \(\(Options[AesBinaryEncrypt] = {CheckFile \[Rule] True};\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "In the binary decryption function, the last byte of the decrypted message \
is checked in the same way as in ",
  StyleBox["AesTextDecrypt",
    FontFamily->"Courier New"],
  ":"
}], "Text"],

Cell[BoxData[
    \(\(AesBinaryDecrypt[cfilename_String, filename_String, key_List, 
          mode_String: "\<cbc\>", opts___Rule] := 
        Module[{b = FileToByteList[cfilename], f = filename, 
            k = CheckKey[key], m = CheckMode[mode], 
            ch = \(CheckFile /. {opts}\) /. \ 
                Options[AesBinaryDecrypt]}, \[IndentingNewLine]If[
            k === Null || b === $Failed, Return[]]; \[IndentingNewLine]If[\ 
            Mod[Length[b], 16] \[NotEqual] 0, 
            Message[AesBinaryDecrypt::notaes, cfilename]; 
            Return[]]; \[IndentingNewLine]If[m === Null, 
            Print["\<Decryption aborted by user\>"]; 
            Return[]]; \[IndentingNewLine]If[ch, f = CheckFileName[f]; 
            If[f === Null, Print["\<Decryption aborted by user\>"]; 
              Return[]]]; \[IndentingNewLine]b = 
            UnpadByteList[
              RijndaelDecryptMode[b, k, m]]; \[IndentingNewLine]If[
            b === Null, Message[AesBinaryDecrypt::badbyte]; Return[]]; 
          ByteListToFile[b, f]];\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(Options[AesBinaryDecrypt] = {CheckFile \[Rule] True};\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "Although we have given functions to encrypt/decrypt both strings and \
files, it may be interesting to have available a function to write a text \
string to a file and another one to read a text string from a file. This will \
be done in binary form so that, for example, the file created by writing a \
string contains exactly that string with nothing else added (this is not the \
case if we use ",
  StyleBox["Write",
    FontFamily->"Courier New"],
  " to write a string to a file). These functions are just a slight \
modification of the functions ",
  StyleBox["ByteListToFile",
    FontFamily->"Courier New"],
  " and ",
  StyleBox["FileToByteList",
    FontFamily->"Courier New"],
  ", defined above."
}], "Text"],

Cell[BoxData[
    \(\(StringToFile[str_String, filename_String] := 
        Module[{f = CheckFileName[filename]}, 
          If[f === Null, Print["\<Operation aborted by user\>"]; Return[], 
            ByteListToFile[ToCharacterCode[str], f]]];\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(FileToString[filename_String] := 
        Module[{b = FileToByteList[filename]}, 
          If[b === $Failed, Return[], FromCharacterCode[b]]];\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "In several of the next functions we want to represent bytes as hexadecimal \
strings. We will give functions that will allow to convert a list of bytes to \
a hexadecimal string and conversely. This will be useful, for example, to \
convert keys supplied by ",
  StyleBox["TextToKey",
    FontFamily->"Courier New"],
  " or ",
  StyleBox["RandomKey",
    FontFamily->"Courier New"],
  " to hexadecimal strings or, conversely, to use a hexadecimal string for \
the key in one of the preceding functions, which require that the key be \
passed to them as a list of bytes.  These functions will also be useful when \
we want to represent the IV or the plaintext as a hex string. We start by \
given a function that converts an integer in the 0-15 range to the \
corresponding hex digit. As hex digits we will use the ten decimal digits in \
the 0-9 range plus the first six lower case letters of the English alphabet."
}], "Text"],

Cell[BoxData[{
    \(\(ToHexDigit[n_Integer?\((0 \[LessEqual] \ # \[LessEqual] \ 15 &)\)] := 
        If[n \[LessEqual] 9, ToString[n], 
          FromCharacterCode[n + 87]];\)\), "\[IndentingNewLine]", 
    \(\(SetAttributes[ToHexDigit, Listable];\)\)}], "Input",
  InitializationCell->True],

Cell["The next function converts a byte to its hex form:", "Text"],

Cell[BoxData[{
    \(\(ByteToHex[
          byte_Integer?\((0 \[LessEqual] \ # \[LessEqual] \ 255 &)\)] := 
        StringJoin[
          ToHexDigit[
            IntegerDigits[byte, 16, 2]]];\)\), "\[IndentingNewLine]", 
    \(\(SetAttributes[ByteToHex, Listable];\)\)}], "Input",
  InitializationCell->True],

Cell["\<\
Now, we give a function to convert a list of bytes to a hexadecimal string:\
\>", "Text"],

Cell[BoxData[
    \(\(BytesToHexString[
          bytes_?\((VectorQ[#, \((0 \[LessEqual] \ # \[LessEqual] \ 
                        255)\) &] &)\)] := 
        StringJoin[ByteToHex[bytes]];\)\)], "Input",
  InitializationCell->True],

Cell["\<\
Later on we will also need the inverse function of the preceding one. First, \
an auxiliary function that converts the code associated with a hex digit to \
the number form of the digit:\
\>", "Text"],

Cell[BoxData[{
    \(\(FromCode[n_Integer] := 
        If[n \[LessEqual] 57, n - 48, n - 87];\)\), "\[IndentingNewLine]", 
    \(\(SetAttributes[FromCode, Listable];\)\)}], "Input",
  InitializationCell->True],

Cell[TextData[{
  "Now, the function to convert a hex string to a list of bytes, namely, the \
inverse function of ",
  StyleBox["BytesToHexString",
    FontFamily->"Courier New"],
  ". Since each byte is represented by two hex digits (with the one on the \
left possibly equal to zero), the input string must actually be a hex string \
of even length, otherwise the last string character is disregarded. This \
function will also accept as hex digits in the input string the upper case \
letters in the A-B range, which will subsequently be converted to lower \
case."
}], "Text"],

Cell[BoxData[
    \(\(HexStringToBytes[str_String] := 
        Module[{b = \(FromDigits[#, 16] &\) /@ 
                Partition[FromCode[ToCharacterCode[ToLowerCase[str]]], 
                  2]}, \[IndentingNewLine]If[\(! VectorQ[
                b, \((0 \[LessEqual] \ # \[LessEqual] \ 255)\) &]\), 
            Message[HexStringToBytes::nohex, str]; Return[], 
            b]];\)\)], "Input",
  InitializationCell->True],

Cell["\<\
The next function is just a variation of the preceding one which will be used \
to convert AES keys given as hexadecimal strings to lists of bytes.  The \
difference is that now we check also the length, returning an error message \
if it does not correspond to a valid key.\
\>", "Text"],

Cell[BoxData[
    \(\(KeyStringToBytes[keystring_String] := 
        Module[{k = HexStringToBytes[keystring]}, \[IndentingNewLine]If[
            k === Null, 
            Return[]]; \[IndentingNewLine]If[\(! MemberQ[{16, 24, 32}, 
                Length[k]]\), Message[KeyStringToBytes::noleng, keystring]; 
            Return[]]; \[IndentingNewLine]k];\)\)], "Input",
  InitializationCell->True],

Cell["\<\
Next we give a function that will provide us with a random key of the \
specified length, given as a hex string. Note that length here refers to the \
key length in bits (128, 192 or 256) and not to the length of the string, \
which will be exactly one fourth of the number of bits.\
\>", "Text"],

Cell[BoxData[
    \(\(RandomKeyString[length_] := 
        Module[{k = RandomKey[length]}, 
          If[k === Null, Return[], BytesToHexString[k]]];\)\)], "Input",
  InitializationCell->True],

Cell["\<\
Similarly, we can obtain a key (as a hex string) from any nonempty text \
string:\
\>", "Text"],

Cell[BoxData[
    \(\(TextToKeyString[text_String, 
          length_] := \[IndentingNewLine]Module[{k = 
              TextToKey[text, length]}, \ 
          If[k === Null, Return[], BytesToHexString[k]]];\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "We are now going to build a function that allows the use of CBC mode for \
authentication purposes, which is made possible by the fact that a change in \
a plaintext block submitted to CBC encryption changes not only the \
corresponding ciphertext block but also all the subsequent blocks. \
Authentication will be obtained by using a MAC (Message Authentication Code) \
which produces, for a given  message, a tag that can be sent attached to it \
in order to guarantee its authenticity. There is some danger of \
terminological confusion here, because the term MAC is commonly used to refer \
both to the authentication scheme and to the tag computed by it, but we hope \
that the precise meaning of the term will be clear from the context. The \
basic CBC-MAC is the last ciphertext block obtained by encrypting the message \
in CBC mode with a public IV and a shared secret key  (cf. ",
  ButtonBox["[2, 10]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  "). The authenticity of the message can be checked by computing the MAC \
through CBC encryption and comparing it with the one received.  The 16-byte \
vector whose components are all zero is used as IV and it is necessary to pad \
the message to obtain a plaintext whose byte length is a multiple of 16. \
Random padding cannot be used in this case, because a reproducible result is \
needed. A possible solution consists in padding with zeros on the right, but \
this is vulnerable to obvious forgeries as all the different messages that \
produce the same padded result would be authenticated by the same MAC. This \
can be avoided by padding with a one-bit followed by zero-bits (the so-called \
",
  Cell[BoxData[
      \(TraditionalForm\`10\^i\)]],
  " padding), which is what we do here, adding a full block in case the \
message length is already a multiple of 16. Observe that, as we work at the \
byte level, we start padding with a byte 128, whose binary form is a 1 \
followed by seven zeros.\n\tThe basic AES CBC-MAC is then given by the \
following function which will be private to this package and only used to \
feed two other functions that will provide the authentication functionality.  \
The output is a list of bytes that can be regarded as the \"basic MAC\". We \
remark that in this function we do not call ",
  StyleBox["AesCbcEncrypt",
    FontFamily->"Courier New"],
  " as we only need the last ciphertext block and it seems more efficient to \
use ",
  StyleBox["Fold",
    FontFamily->"Courier New"],
  " (instead of ",
  StyleBox["FoldList",
    FontFamily->"Courier New"],
  "), which gives us precisely what is required, without having to drop \
anything:"
}], "Text"],

Cell[BoxData[
    \(\(RijndaelBasicCbcMac[bytes_List, \ key_List] := 
        Module[{b = Join[bytes, {128}], l, 
            iv = Table[0, {16}]}, \[IndentingNewLine]l = Length[b]; 
          If[\ Mod[l, 16] \[NotEqual] 0, 
            b = PadRight[b, 
                l + 16 - Mod[l, 16]]]; \[IndentingNewLine]\ \ \ \ b = 
            Partition[Join[iv, b], 16]; \ \ \ \ \ \ \[IndentingNewLine]Fold[
            RijndaelEncrypt[BitXor[#1, #2], key] &, First[b], 
            Rest[b]]];\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "The authentication provided by the preceding function is secure if only \
fixed length messages are used but it is vulnerable to forgeries in the case \
of variable length messages (cf. ",
  ButtonBox["[2, 10]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  "). For this reason, an \"output transformation\" is usually applied to the \
basic MAC thus produced, obtaining in this way different variants of the \
AES-based CBC-MAC. A simple possibility consists in truncating the output of \
the preceding function (i.e., the last ciphertext block) to a number of bits \
",
  StyleBox["m",
    FontSlant->"Italic"],
  " < 128 as in the next function, which produces a MAC consisting of the 8 \
first bytes of the basic MAC.  For convenience reasons, the output MAC is \
given as a hex string, which is more compact than a list of bytes in decimal \
notation. To achieve some uniformity between input and output, the key in \
this case will also be given as a hex string that can be obtained, for \
example, using ",
  StyleBox["RandomKeyString",
    FontFamily->"Courier New"],
  ", whose use is advisable because of its pseudo-random character and also \
because it makes unnecessary to type the key at the keyboard. Of course, it \
is easy to convert between lists of bytes and hex strings, and conversely, by \
using ",
  StyleBox["BytesToHexString",
    FontFamily->"Courier New"],
  " and ",
  StyleBox["HexStringToBytes",
    FontFamily->"Courier New"],
  StyleBox[" ",
    FontFamily->"Times New Roman"],
  StyleBox["(or",
    FontFamily->"Times New Roman"],
  StyleBox[" KeyStringToBytes)",
    FontFamily->"Courier New"],
  ". Apart from the key, the function requires as input the name of the file \
which is going to be authenticated."
}], "Text"],

Cell[BoxData[
    \(\(AesCbcMac[filename_String, \ \ keystring_String] := 
        Module[{k = KeyStringToBytes[keystring], 
            b = FileToByteList[filename], m}, \[IndentingNewLine]If[
            k === Null || b === $Failed, \ Return[]]; \[IndentingNewLine]m = 
            RijndaelBasicCbcMac[b, k]; \[IndentingNewLine]BytesToHexString[
            Take[m, 8]]];\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "A more secure (albeit slower) alternative to the preceding authentication \
method consists in using Rijndael encryption as \"output transformation\", \
i.e., in encrypting the basic AES CBC-MAC with a different key. This method \
is often referred to as EMAC (for \"encrypted MAC\") and it has been shown to \
be secure with inputs of arbitrary length ",
  ButtonBox["[10]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  ". It is given by the next function, in which the input is the name of the \
file to be authenticated and the keys, which should be supplied as hex \
strings. The output MAC is also a hex string:"
}], "Text"],

Cell[BoxData[
    \(\(AesCbcEmac[filename_String, \ keystring1_String, \ 
          keystring2_String] := 
        Module[{k1 = KeyStringToBytes[keystring1], 
            k2 = KeyStringToBytes[keystring2], b = FileToByteList[filename], 
            m}, \[IndentingNewLine]If[
            k1 === Null || k2 === Null || b === $Failed, 
            Return[]]; \[IndentingNewLine]m = 
            RijndaelBasicCbcMac[b, k1]; \[IndentingNewLine]BytesToHexString[
            RijndaelEncrypt[m, k2]]];\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "The last variant of CBC-MAC which we are going to implement is OMAC \
(\"One-key CBC-MAC\"), which is the MAC algorithm that NIST intends to \
specify as standard for AES-based message authentication. Like EMAC, OMAC has \
also been proven to be secure and it has better performance characteristics \
as it only requires one key scheduling (instead of two required by EMAC). \
OMAC, designed by T. Iwata and K. Kurosawa ",
  ButtonBox["[6]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  " is a variation of the XCBC algorithm that had been proposed by Black and \
Rogaway. In XCBC (we will always refer to the AES version of the MAC \
algorithms; they can also be used with other block ciphers as well) the \
padding algorithm used for the basic CBC-MAC is modified by not adding an \
extra block when the length of the message in bits is already a nonzero \
multiple of 128. In addition to an AES key (",
  StyleBox["k",
    FontSlant->"Italic"],
  Cell[BoxData[
      \(TraditionalForm\`\_1\)],
    FontSlant->"Italic"],
  ", say), two additional 128-bit whitening keys, ",
  StyleBox["k",
    FontSlant->"Italic"],
  Cell[BoxData[
      \(TraditionalForm\`\_2\)],
    FontSlant->"Italic"],
  ", ",
  StyleBox["k",
    FontSlant->"Italic"],
  Cell[BoxData[
      \(TraditionalForm\`\_3\)],
    FontSlant->"Italic"],
  ", are used. If no padding was added, the algorithm is the same as CBC-MAC \
except for the fact that, before encrypting the last block with ",
  StyleBox["k",
    FontSlant->"Italic"],
  Cell[BoxData[
      \(TraditionalForm\`\_1\)],
    FontSlant->"Italic"],
  ", this block is XOR-ed with key ",
  StyleBox["k",
    FontSlant->"Italic"],
  Cell[BoxData[
      \(TraditionalForm\`\_2\)],
    FontSlant->"Italic"],
  ". If padding was added, a similar operation is performed, except for the \
fact that the last block is XOR-ed with ",
  StyleBox["k",
    FontSlant->"Italic"],
  Cell[BoxData[
      \(TraditionalForm\`\_3\)],
    FontSlant->"Italic"],
  " before encryption. The OMAC variation of XCBC reduces the number of keys \
to one (the AES key ",
  StyleBox["k",
    FontSlant->"Italic"],
  Cell[BoxData[
      \(TraditionalForm\`\_1\)],
    FontSlant->"Italic"],
  " = ",
  StyleBox["k",
    FontSlant->"Italic"],
  ") by deriving ",
  StyleBox["k",
    FontSlant->"Italic"],
  Cell[BoxData[
      \(TraditionalForm\`\_2\)],
    FontSlant->"Italic"],
  " and ",
  StyleBox["k",
    FontSlant->"Italic"],
  Cell[BoxData[
      \(TraditionalForm\`\_3\)],
    FontSlant->"Italic"],
  " from ",
  StyleBox["k",
    FontSlant->"Italic"],
  ". This is done as follows. Let ",
  StyleBox["L",
    FontSlant->"Italic"],
  " = ",
  Cell[BoxData[
      \(TraditionalForm\`E\_k\)]],
  "(",
  Cell[BoxData[
      \(TraditionalForm\`0\^128\)]],
  "), where ",
  Cell[BoxData[
      \(TraditionalForm\`E\_k\)]],
  " denotes Rijndael encryption with key ",
  StyleBox["k",
    FontSlant->"Italic"],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`0\^128\)]],
  " is the block with all 128 bits equal to zero. Consider the field GF(",
  Cell[BoxData[
      \(TraditionalForm\`2\^128\)]],
  "), whose elements can be represented, in a way similar to the one used in \
the case of  GF(",
  Cell[BoxData[
      \(TraditionalForm\`2\^8\)]],
  "), by the numbers in the [0, ",
  Cell[BoxData[
      \(TraditionalForm\`2\^128\)]],
  "-1] range and hence also by the 128-bit (or 16-byte) vectors. Then take ",
  StyleBox["k",
    FontSlant->"Italic"],
  Cell[BoxData[
      \(TraditionalForm\`\_2\)],
    FontSlant->"Italic"],
  " = 2 \[Bullet] ",
  StyleBox["L",
    FontSlant->"Italic"],
  ", and ",
  StyleBox["k",
    FontSlant->"Italic"],
  Cell[BoxData[
      \(TraditionalForm\`\_3\)],
    FontSlant->"Italic"],
  " = 4 \[Bullet] ",
  StyleBox["L",
    FontSlant->"Italic"],
  ", where the multiplication is in GF(",
  Cell[BoxData[
      \(TraditionalForm\`2\^128\)]],
  ") and proceed exactly as in XCBC. This is the OMAC1 variant, which NIST \
intends to standardize under the name CMAC ",
  ButtonBox["[10]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  ". The other variant of OMAC, called OMAC2, is obtained in a similar manner \
but taking instead ",
  StyleBox["k",
    FontSlant->"Italic"],
  Cell[BoxData[
      \(TraditionalForm\`\_3\)],
    FontSlant->"Italic"],
  " = ",
  Cell[BoxData[
      \(TraditionalForm\`2\^\(-1\)\)]],
  "\[Bullet] ",
  StyleBox["L",
    FontSlant->"Italic"],
  " (where ",
  Cell[BoxData[
      \(TraditionalForm\`2\^\(-1\)\)]],
  " is, of course, the inverse of 2 in GF(",
  Cell[BoxData[
      \(TraditionalForm\`2\^128\)]],
  ") which, just for the record, is equal to ",
  Cell[BoxData[
      \(TraditionalForm\`2\^127\)]],
  "+ 67 ).\n\tTo implement OMAC authentication, we start by defining the \
multiplication and division by 2 in GF(",
  Cell[BoxData[
      \(TraditionalForm\`2\^128\)]],
  "). Of course, these are particular cases of the field multiplication and \
so we could just invoke the package ",
  StyleBox["Algebra`FiniteFields` ",
    FontFamily->"Courier New"],
  StyleBox["but we will implement these operations directly as we did in the \
field ",
    FontFamily->"Times New Roman"],
  "GF(",
  Cell[BoxData[
      \(TraditionalForm\`2\^8\)]],
  "), using only a shift and a ",
  StyleBox["BitXor",
    FontFamily->"Courier New"],
  StyleBox[" operation in each case. Now, the field elements are given by \
lists of bytes instead of integers and so we start by giving a couple of \
functions which take such a list and perform a cyclic bitwise left and a \
cyclic bitwise right shift, respectively:",
    FontFamily->"Times New Roman"]
}], "Text"],

Cell[BoxData[
    \(\(BitRotateLeft[bytes_List] := \(FromDigits[#, 2] &\) /@ 
          Partition[RotateLeft[Flatten[IntegerDigits[bytes, 2, 8]]], 
            8];\)\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(\(BitRotateRight[bytes_List] := \(FromDigits[#, 2] &\) /@ 
          Partition[RotateRight[Flatten[IntegerDigits[bytes, 2, 8]]], 
            8];\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "The irreducible polynomial used in OMAC to define the GF(",
  Cell[BoxData[
      \(TraditionalForm\`2\^128\)]],
  ") arithmetic is ",
  Cell[BoxData[
      \(TraditionalForm\`X\^128\)]],
  "+ ",
  Cell[BoxData[
      \(TraditionalForm\`X\^7\)]],
  "+ ",
  Cell[BoxData[
      \(TraditionalForm\`X\^2\)]],
  "+ ",
  Cell[BoxData[
      \(TraditionalForm\`X\)]],
  " + 1. Taking this into account it is easy to see that multiplication and \
division by 2, respectively, are given by (cf. ",
  ButtonBox["[6]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  "):"
}], "Text"],

Cell[BoxData[
    \(\(MultBy2[bytes_List] := 
        Module[{f = First[bytes], r = BitRotateLeft[bytes]}, 
          If[f < 128, r, 
            BitXor[r, Join[Table[0, {15}], {134}]]]];\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "This function is similar to ",
  StyleBox["ByteTimes2",
    FontFamily->"Courier New"],
  StyleBox[",",
    FontFamily->"Times New Roman"],
  " which we had previously defined to multiply by 2 in the byte space,  but \
the situation is slightly different now because we are using a cyclic shift \
(i.e., a rotation of bits) instead of just appending a 0-bit at the end. We \
now xor with 134 because the irreducible polynomial used to define the \
multiplication in GF(",
  Cell[BoxData[
      \(TraditionalForm\`2\^128\)]],
  ") is  ",
  Cell[BoxData[
      \(TraditionalForm\`X\^128\)]],
  "+ ( ",
  Cell[BoxData[
      \(TraditionalForm\`X\^7\)]],
  "+ ",
  Cell[BoxData[
      \(TraditionalForm\`X\^2\)]],
  "+ ",
  Cell[BoxData[
      \(TraditionalForm\`X\)]],
  ") + 1, and the integer corresponding to ",
  Cell[BoxData[
      \(TraditionalForm\`X\^7\)]],
  "+ ",
  Cell[BoxData[
      \(TraditionalForm\`X\^2\)]],
  "+ ",
  Cell[BoxData[
      \(TraditionalForm\`X\)]],
  " is, precisely, 134. For the inverse function, we could just use, when the \
integer to be multiplied is odd (otherwise, only a right shift is required), \
the inverse operations in reverse order (first ",
  StyleBox["BitXor",
    FontFamily->"Courier New"],
  ", then right shift). However, division by 2 can also be implemented in \
this case by making first a bitwise right shift and then a ",
  StyleBox["BitXor",
    FontFamily->"Courier New"],
  " as before. Since the result of applying a bitwise right shift to 134 \
(i.e., of dividing it by 2) is precisely 67, this can be accomplished by a \
right shift followed by xoring with 67. "
}], "Text"],

Cell[BoxData[
    \(\(DivBy2[bytes_List] := \ 
        Module[{l = Last[bytes], 
            r = BitRotateRight[bytes]}, \[IndentingNewLine]If[EvenQ[l], r, 
            BitXor[r, Join[Table[0, {15}], {67}]]]];\)\)], "Input",
  InitializationCell->True],

Cell["\<\
The next function provides Rijndael-based OMAC authentication. This function \
takes as input the message and the key, both of them given as lists of bytes, \
and the type, which can be either 1 or 2 according to whether OMAC1 or OMAC2 \
authentication is required. This function will only be used when called from \
other functions to be defined later, which will provide authentication for \
files and message strings.\
\>", "Text"],

Cell[BoxData[
    \(\(RijndaelOmac[bytes_List, key_List, 
          type_Integer] := \[IndentingNewLine]Module[{b = bytes, k = key, 
            fullblock = False, iv = Table[0, {16}], l, lb, 
            x}, \[IndentingNewLine]l = Length[b]; 
          If[\ Mod[l, 16] \[NotEqual] 0 || l \[Equal] 0, b = Join[b, {128}]; 
            b = PadRight[b, l + 16 - Mod[l, 16]], fullblock = True]; 
          b = Partition[Join[iv, b], 16]; 
          lb = Last[b]; \ \ x = RijndaelEncrypt[iv, k]; \[IndentingNewLine]If[
            fullblock, b = ReplacePart[b, BitXor[lb, MultBy2[x]], \(-1\)], 
            If[type \[Equal] 1, 
              b = ReplacePart[b, BitXor[lb, MultBy2[MultBy2[x]]], \(-1\)], \ 
              b = ReplacePart[b, 
                  BitXor[lb, DivBy2[x]]\ , \(-1\)]]]; \[IndentingNewLine]Fold[
            RijndaelEncrypt[BitXor[#1, #2], k] &, First[b], 
            Rest[b]]];\)\)], "Input",
  InitializationCell->True],

Cell["\<\
Now we can give the function that produces the OMAC associated to a file. The \
input arguments are the file name, the key and the type. As before, the key \
is supplied as a hex string and the output MAC will also be a hex string. The \
type (either 1, for OMAC1, or 2, for OMAC2) is supplied to the function as \
its last argument; if no type is specified, then OMAC1 is used by default.\
\>", "Text"],

Cell[BoxData[
    \(\(AesOmac[filename_String, \ keystring_String, type_:  1] := 
        Module[{b = FileToByteList[filename], 
            k = KeyStringToBytes[keystring]}, \[IndentingNewLine]If[
            k === Null || b === $Failed, 
            Return[]]; \[IndentingNewLine]If[\(! MemberQ[{1, 2}, type]\), 
            Message[AesOmac::notype]; 
            Return[]]; \[IndentingNewLine]BytesToHexString[
            RijndaelOmac[b, k, type]]];\)\)], "Input",
  InitializationCell->True],

Cell["\<\
We also give a function that computes the OMAC for text strings:\
\>", "Text"],

Cell[BoxData[
    \(\(AesTextOmac[text_String, \ keystring_String, type_:  1] := 
        Module[{b = ToCharacterCode[text], 
            k = KeyStringToBytes[keystring]}, \[IndentingNewLine]If[
            k === Null, 
            Return[]]; \[IndentingNewLine]If[\(! MemberQ[{1, 2}, type]\), 
            Message[AesOmac::notype]; 
            Return[]]; \[IndentingNewLine]BytesToHexString[
            RijndaelOmac[b, k, type]]];\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "In relation with the two preceding OMAC functions we remark that, if we \
write a text string to a file using, for example, ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  "'s function ",
  StyleBox["Write",
    FontFamily->"Courier New"],
  ", the tag for this file computed by ",
  StyleBox["AesOmac",
    FontFamily->"Courier New"],
  " will be different from the tag for the text string computed by ",
  StyleBox["AesTextOmac",
    FontFamily->"Courier New"],
  ". The reason is that the file contains some additional bytes which do not \
belong to the text string (the quotation marks and a CR/LF at the end if ",
  StyleBox["Write",
    FontFamily->"Courier New"],
  " was used). Since the file is read by ",
  StyleBox["AesOmac",
    FontFamily->"Courier New"],
  " in binary form, these extra bytes are taken into account when computing \
the MAC. However, if we edit the file to remove the extra bytes, the \
resulting MAC's will be identical. The MAC's will also be the same if we \
write or read the text string using ",
  StyleBox["StringToFile",
    FontFamily->"Courier New"],
  " or ",
  StyleBox["FileToString",
    FontFamily->"Courier New"],
  "."
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Some tests", "Section"],

Cell[TextData[{
  "Next we provide some tools to test the AES code in this notebook. We will \
start by setting up an explicit display of some of the constants and tables \
used in Rijndael, where the bytes will be given in hex form. This will be \
done with the help of the next function, whose argument is a string that can \
take the values \"Sbox\" (to display the ",
  StyleBox["S Box",
    FontSlant->"Italic"],
  "), \"ISbox\" (the inverse ",
  StyleBox["S Box",
    FontSlant->"Italic"],
  "), \"MC\" (the matrix used by ",
  StyleBox["MixColumns",
    FontFamily->"Courier New"],
  "), \"IMC\" (the inverse of the previous matrix), and \"Rcon\" (the round \
constants)."
}], "Text"],

Cell[BoxData[
    \(\(DisplayRijndaelObject[
          objectname_String] \
:= \[IndentingNewLine]Which[\[IndentingNewLine]objectname == "\<Sbox\>", \
\[IndentingNewLine]\ \ TableForm[ByteToHex[Partition[S, 16]], \ 
            TableHeadings -> ToHexDigit[{Range[16] - 1, Range[16] - 1}], 
            TableSpacing \[Rule] {1, 
                1}], \[IndentingNewLine]objectname == "\<ISbox\>", \
\[IndentingNewLine]\ \ TableForm[ByteToHex[Partition[IS, 16]], \ 
            TableHeadings -> ToHexDigit[{Range[16] - 1, Range[16] - 1}], 
            TableSpacing \[Rule] {1, 
                1}], \[IndentingNewLine]objectname == "\<MC\>", \
\[IndentingNewLine]\ \ \ MatrixForm[
            ByteToHex[
              MC]], \[IndentingNewLine]objectname == "\<IMC\>", \
\[IndentingNewLine]\ \ \ MatrixForm[
            ByteToHex[
              IMC]], \[IndentingNewLine]objectname == "\<Rcon\>", \
\[IndentingNewLine]\ \ \ TableForm[ByteToHex[Rcon], 
            TableHeadings \[Rule] {Automatic, 
                None}], \[IndentingNewLine]True, \[IndentingNewLine]\ \ \ \
Message[DisplayRijndaelObject::noobj]\[IndentingNewLine]];\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "Next, we introduce the KAT's (\"Known Answer Tests\") defined by  NIST to \
validate candidate algorithm submissions. The KAT vectors obtained from these \
tests output are available at the NIST Web page: ",
  ButtonBox["http://csrc.nist.gov/cryptval/",
    ButtonData:>{
      URL[ "http://csrc.nist.gov/cryptval/"], None},
    ButtonStyle->"Hyperlink"],
  " and in ",
  ButtonBox["[1]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  ".  The output of the next test programs should be compared with these \
values.\n\tWe start with the ECB Variable Text KAT's.  It is perhaps \
worthwhile to mention that the input plaintexts are, when regarded as 128-bit \
vectors, the sums of the first ",
  StyleBox["i",
    FontSlant->"Italic"],
  " unitary vectors, with ",
  StyleBox["i",
    FontSlant->"Italic"],
  " ranging from 1 to 128 (instead of just the unitary vectors taken as input \
in other versions of the test which, of course, produce a different output). \
A similar remark applies to the Variable Key Texts below. The tests are given \
as functions of the keylength, which can only take as values 128, 192, or \
256, and the output ciphertexts are printed as hex strings of bytes."
}], "Text"],

Cell[BoxData[
    \(\(VarTextTest[length_?\((MemberQ[{128, 192, 256}, #] &)\)] := 
        Module[{k = Table[0, {length/8}], 
            t = \(FromDigits[#, 2] &\) /@ 
                Partition[
                  Flatten[
                    NestList[ReplacePart[RotateRight[#], 1, 1] &, 
                      Join[{1}, Table[0, {127}]], 127]], 
                  8]}, \ \[IndentingNewLine]t = 
            RijndaelEncrypt[t, k]; \[IndentingNewLine]Do[
            Print[i, "\<: \>", 
              BytesToHexString[\(Partition[t, 16]\)[\([i]\)]]], {i, 
              128}]];\)\)], "Input",
  InitializationCell->True],

Cell["\<\
The next function first computes the input of the Variable Key tests and then \
produces the output through Rijndael encryption:\
\>", "Text"],

Cell[BoxData[
    \(\(VarKeyTest[length_?\((MemberQ[{128, 192, 256}, #] &)\)] := 
        Module[{bkl = \ \(Partition[#, 8] &\) /@ 
                NestList[ReplacePart[RotateRight[#], 1, 1] &, 
                  Join[{1}, Table[0, {length - 1}]], length - 1], 
            pt = Table[0, {16}]}, \[IndentingNewLine]Do[
            Print[i, \ "\<: \>", 
              BytesToHexString[
                RijndaelEncrypt[
                  pt, \(FromDigits[#, 2] &\) /@ bkl[\([i]\)]]]], {i, 
              length}]];\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "Next we are going to give tests for the encryption and decryption \
functions in all supported modes. The plaintexts, the keys, and the IV's, \
will be supplied to these tests in the form of hex strings. The code to test \
CBC and CFB encryption and decryption is the same as in the preceding \
encryption and decryption functions except for the fact that",
  StyleBox[", in this case, the IV is also supplied to the function instead \
of being (pseudo) randomly generated. Moreover, the ciphertext produced by \
the next function does not include (to avoid some redundancy) the IV as is \
the case with the ciphertext obtained from ",
    FontFamily->"Times New Roman"],
  StyleBox["CbcEncrypt",
    FontFamily->"Courier New"],
  StyleBox[" and ",
    FontFamily->"Times New Roman"],
  StyleBox["CfbEncrypt",
    FontFamily->"Courier New"],
  StyleBox[". If no IV is specified, the string corresponding to the first 16 \
bytes in ascending order is used by default. This function has no default \
mode and ECB mode can also be specified, in which case the IV is not needed.",
    
    FontFamily->"Times New Roman"]
}], "Text"],

Cell[BoxData[
    \(\(AesEncryptTest[plaintext_String, keystring_String, mode_String, 
          iv_String: 
             BytesToHexString[
              Range[16] - 1]] := \[IndentingNewLine]Module[{b = 
              HexStringToBytes[plaintext], k = KeyStringToBytes[keystring], 
            i = HexStringToBytes[iv], 
            m = CheckMode[mode]}, \[IndentingNewLine]If[
            k === Null || i === Null, \ Return[]]; \[IndentingNewLine]If[
            m === Null, Print["\<Test aborted by user\>"]; 
            Return[]]; \[IndentingNewLine]If[Length[i] \[NotEqual] 16, 
            Message[AesEncryptTest::noleng, iv]; 
            Return[]]; \[IndentingNewLine]b = 
            BytesToHexString[RijndaelEncryptMode[b, k, m, i]]; 
          If[m \[Equal] "\<ecb\>", b, StringDrop[b, 32]]];\)\)], "Input",
  InitializationCell->True],

Cell["\<\
The next  function tests decryption in an entirely  similar way:\
\>", "Text"],

Cell[BoxData[
    \(\(AesDecryptTest[ciphertext_String, keystring_String, mode_String, 
          iv_String: 
             BytesToHexString[
              Range[16] - 1]] := \[IndentingNewLine]Module[{b = 
              HexStringToBytes[ciphertext], k = KeyStringToBytes[keystring], 
            i = HexStringToBytes[iv], 
            m = CheckMode[mode]}, \[IndentingNewLine]If[
            k === Null || i === Null, \ Return[]]; \[IndentingNewLine]If[
            m === Null, Print["\<Test aborted by user\>"]; 
            Return[]]; \[IndentingNewLine]If[Length[i] \[NotEqual] 16, 
            Message[AesDecryptTest::noleng, iv]; 
            Return[]]; \[IndentingNewLine]If[m != "\<ecb\>", 
            b = Join[i, b]]; \[IndentingNewLine]BytesToHexString[
            RijndaelDecryptMode[b, k, m]]];\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "The next couple of functions do not check compliance with the standard \
but, rather, the fact that the encryption and decryption functions are mutual \
inverses, and the behavior of the input-output functions. The next function \
is a random test which  generates random text strings of variable length and, \
for each of these strings, a random Rijndael key of any one of the three \
possible keylengths. Moreover, the mode to be used is also randomly chosen \
between the three possible ones. This test submits each text string to a \
Rijndael encryption followed by a decryption using the same key and mode, and \
compares the result with the original string; it is successfully passed when \
all these comparisons match, otherwise a message is printed along with the \
string, the key, and the mode for which the test failed. The function takes \
as input the number ",
  StyleBox["n",
    FontFamily->"Courier New"],
  " of encryption-decryption stages to be performed, and the maximum length \
",
  StyleBox["l,",
    FontFamily->"Courier New"],
  " in bytes, of the strings to be generated. It is advisable not to use very \
high values for both parameters in order to let the test be finished in a \
reasonable time."
}], "Text"],

Cell[BoxData[
    \(\(AesRandomTest[n_Integer?Positive, l_Integer?Positive] := 
        Module[{t, k, j, i = 1, m = {"\<ecb\>", "\<cbc\>", "\<cfb\>"}, 
            cond = True}, 
          While[cond\  && i \[LessEqual] n, \ \[IndentingNewLine]k = 
              RandomKey[64\ Random[Integer, {2, 4}]]; 
            t = FromCharacterCode[
                Table[Random[
                    Integer, {32, 255}], {Random[
                      Integer, {1, l}]}]]; \[IndentingNewLine]j = 
              Random[Integer, {1, 
                  3}]; \[IndentingNewLine]cond = \((AesTextDecrypt[
                    AesTextEncrypt[t, k, m[\([j]\)]], k, m[\([j]\)]] \[Equal] 
                  t)\); If[Mod[i, 100] \[Equal] 0, \ 
              Print[i, \ "\< tests\>"]]; \(i++\)]; 
          If[cond, Print[n, \ "\< tests successfully passed\>"], \ 
            Print["\<Test failed at plaintext: \r\n\>", \ 
              t, "\<\r\n with key: \r\n\>", k, \ "\<\r\n in \>", \ 
              ToUpperCase[m[\([j]\)]], "\< mode.\>"]]];\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "Next we give a function to test also the binary input-output routines. For \
this we will pick one file in the current directory (or any file whose name \
is given with the full path) and submit it to a repeated \
encryption-decryption process in the specified mode (CBC mode, by default) \
using,  for each encryption-decryption, a randomly chosen key. The final \
result is compared to the original file and, if they match, the test is \
successfully passed. Of course, since the original file will be overwritten \
several times, this test should only be applied to easily replaceable files \
or to files that can be discarded. The input is the name of the file and the \
number of encryption-decryption stages to be performed. Note that, to avoid \
prompting the user about whether the file should be overwritten, the ",
  StyleBox["CheckFile",
    FontFamily->"Courier New"],
  " option of ",
  StyleBox["AesBinaryEncrypt",
    FontFamily->"Courier New"],
  " and ",
  StyleBox["AesBinaryDecrypt",
    FontFamily->"Courier New"],
  " is set to ",
  StyleBox["False",
    FontFamily->"Courier New"],
  "."
}], "Text"],

Cell[BoxData[
    \(\(AesRandomBinaryTest[filename_String, mode_String: "\<cbc\>", 
          n_Integer?Positive] := 
        Module[{b = FileToByteList[filename], c, k, 
            m = CheckMode[mode]}, \[IndentingNewLine]If[b === $Failed, 
            Return[]]; \[IndentingNewLine]If[m === Null, 
            Print["\<Test aborted by user\>"]; 
            Return[]]; \[IndentingNewLine]\ 
          Do[k = RandomKey[
                64\ Random[
                    Integer, {2, 4}]]; \[IndentingNewLine]AesBinaryEncrypt[
              filename, filename, k, m, 
              CheckFile \[Rule] False]; \[IndentingNewLine]AesBinaryDecrypt[
              filename, filename, k, m, CheckFile \[Rule] False], {n}]; 
          c = FileToByteList[filename]; \[IndentingNewLine]If[
            b \[Equal] 
              c, \[IndentingNewLine]Print["\<Test successfully passed\>"], \ 
            Print["\<Test failed\>"]]];\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "Finally, we give a version of OMAC for hexadecimal text strings, which can \
be used as a test of the OMAC implementation. The input is the message as a \
hexadecimal string, the key, also in this format, and the type (1 or 2, type \
1 is used by default). The output is the MAC associated to the message, given \
also as a hex string. The OMAC test vectors can be obtained via the OMAC page \
",
  ButtonBox["[13]",
    ButtonData:>"Refs",
    ButtonStyle->"Hyperlink"],
  ", at the links  ",
  ButtonBox["http://crypt.cis.ibaraki.ac.jp/omac/docs/omac-ad.pdf",
    ButtonData:>{
      URL[ "http://crypt.cis.ibaraki.ac.jp/omac/docs/omac-ad.pdf"], None},
    ButtonStyle->"Hyperlink"],
  " for OMAC1 and ",
  ButtonBox["http://crypt.cis.ibaraki.ac.jp/omac/docs/omac-tv.pdf",
    ButtonData:>{
      URL[ "http://crypt.cis.ibaraki.ac.jp/omac/docs/omac-tv.pdf"], None},
    ButtonStyle->"Hyperlink"],
  "  for OMAC2."
}], "Text"],

Cell[BoxData[
    \(\(OmacTest[text_String, \ keystring_String, \ type_:  1] := 
        Module[{b = HexStringToBytes[text], 
            k = KeyStringToBytes[keystring]}, \[IndentingNewLine]If[
            b === Null || k === Null, 
            Return[]]; \[IndentingNewLine]If[\(! MemberQ[{1, 2}, type]\), 
            Message[OmacTest::notype]; 
            Return[]]; \[IndentingNewLine]BytesToHexString[
            RijndaelOmac[b, k, type]]];\)\)], "Input",
  InitializationCell->True]
}, Closed]],

Cell[CellGroupData[{

Cell["References", "Section"],

Cell[TextData[{
  "[1] L.E. Bassham III, The Advanced Encryption Standard Algorithm \
Validation Suite (AESAVS), NIST publication available at ",
  ButtonBox["http://csrc.nist.gov/cryptval/aes/AESAVS.pdf",
    ButtonData:>{
      URL[ "http://csrc.nist.gov/cryptval/aes/AESAVS.pdf"], None},
    ButtonStyle->"Hyperlink"],
  "\n[2] M. Bellare, P. Rogaway, Introduction to Modern Cryptography, \
available via ",
  ButtonBox["http://www-cse.ucsd.edu/users/mihir/",
    ButtonData:>{
      URL[ "http://www-cse.ucsd.edu/users/mihir/"], None},
    ButtonStyle->"Hyperlink"],
  "\n[3] J. Daemen, V. Rijmen, AES Proposal: Rijndael, available via ",
  ButtonBox["http://csrc.nist.gov/CryptoToolkit/aes/rijndael/",
    ButtonData:>{
      URL[ "http://csrc.nist.gov/CryptoToolkit/aes/rijndael/"], None},
    ButtonStyle->"Hyperlink"],
  "\n[4] J. Daemen, V. Rijmen, The design of Rijndael, Springer-Verlag, \
Berlin (2002).\n[5] M. Dworkin, Recommendation for Block Cipher Modes of \
Operation, NIST Special Publication 800-38A, 2001 Edition, available via ",
  ButtonBox["http://csrc.nist.gov/CryptoToolkit/modes/",
    ButtonData:>{
      URL[ "http://csrc.nist.gov/CryptoToolkit/modes/"], None},
    ButtonStyle->"Hyperlink"],
  "\n[6] T. Iwata, K. Kurosawa, OMAC: One-Key CBC MAC, available via ",
  ButtonBox["http://crypt.cis.ibaraki.ac.jp/omac/docs/omac.pdf",
    ButtonData:>{
      URL[ "http://crypt.cis.ibaraki.ac.jp/omac/docs/omac.pdf"], None},
    ButtonStyle->"Hyperlink"],
  "\n[7] L.R. Knudsen,  Block Ciphers,  in \"Encyclopedia of Cryptography and \
Security\", available via ",
  ButtonBox["http://www.win.tue.nl/~henkvt/Content.html",
    ButtonData:>{
      URL[ "http://www.win.tue.nl/~henkvt/Content.html"], None},
    ButtonStyle->"Hyperlink"],
  "\n[8] H.W. Lenstra, Jr., Rijndael for algebraists, available via ",
  ButtonBox["http://math.berkeley.edu/~hwl/",
    ButtonData:>{
      URL[ "http://math.berkeley.edu/~hwl/"], None},
    ButtonStyle->"Hyperlink"],
  "\n[9] R. Lidl, H. Niederreiter, Introduction to finite fields and their \
applications, Cambridge University Press, Cambridge (1994).\n[10] B. Preneel, \
CBC-MAC and variants,  in \"Encyclopedia of Cryptography and Security\", \
available via ",
  ButtonBox["http://www.win.tue.nl/~henkvt/Content.html",
    ButtonData:>{
      URL[ "http://www.win.tue.nl/~henkvt/Content.html"], None},
    ButtonStyle->"Hyperlink"],
  "\n[11] Specification for the Advanced Encryption Standard (AES), FIPS \
publication 197 (2001), available via ",
  ButtonBox["http://csrc.nist.gov/CryptoToolkit/aes/rijndael/",
    ButtonData:>{
      URL[ "http://csrc.nist.gov/CryptoToolkit/aes/rijndael/"], None},
    ButtonStyle->"Hyperlink"],
  "\n[12] D.R. Stinson, Cryptography, Theory and Practice, Second Edition, \
Chapman & Hall/CRC, Boca Raton (2002).\n[13] The OMAC page, ",
  ButtonBox["http://crypt.cis.ibaraki.ac.jp/omac/omac.html",
    ButtonData:>{
      URL[ "http://crypt.cis.ibaraki.ac.jp/omac/omac.html"], None},
    ButtonStyle->"Hyperlink"],
  "\n[14] The Rijndael Page, ",
  ButtonBox["http://www.esat.kuleuven.ac.be/~rijmen/rijndael/",
    ButtonData:>{
      URL[ "http://www.esat.kuleuven.ac.be/~rijmen/rijndael/"], None},
    ButtonStyle->"Hyperlink"],
  "\n[15] W. Trappe, L.C. Washington, Introduction to Cryptography with \
Coding Theory, Prentice Hall, Upper Saddle River (2002).\n[16] D. Wagner, \
Towards a unifying view of block cipher cryptanalysis, available via ",
  ButtonBox["http://www.cs.berkeley.edu/~daw/papers/",
    ButtonData:>{
      URL[ "http://www.cs.berkeley.edu/~daw/papers/"], None},
    ButtonStyle->"Hyperlink"]
}], "Text",
  CellTags->"Refs"]
}, Closed]],

Cell[CellGroupData[{

Cell["The End", "Section",
  Evaluatable->False,
  AspectRatioFixed->False],

Cell["End[ ];", "Input",
  InitializationCell->True,
  AspectRatioFixed->False],

Cell["EndPackage[ ];", "Input",
  InitializationCell->True,
  AspectRatioFixed->False]
}, Closed]]
}, Closed]]
},
FrontEndVersion->"5.1 for Microsoft Windows",
ScreenRectangle->{{0, 1280}, {0, 951}},
AutoGeneratedPackage->None,
WindowSize->{787, 512},
WindowMargins->{{Automatic, 100}, {Automatic, 47}},
PrintingCopies->1,
PrintingPageRange->{Automatic, Automatic},
CellLabelAutoDelete->True
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{
  "Refs"->{
    Cell[124226, 3117, 3653, 74, 413, "Text",
      CellTags->"Refs"]}
  }
*)

(*CellTagsIndex
CellTagsIndex->{
  {"Refs", 128898, 3229}
  }
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 49, 0, 95, "Title"],

Cell[CellGroupData[{
Cell[1850, 57, 25, 0, 73, "Section"],
Cell[1878, 59, 416, 13, 112, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[2331, 77, 35, 0, 43, "Section"],
Cell[2369, 79, 661, 12, 128, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[3067, 96, 26, 0, 43, "Section"],
Cell[3096, 98, 27, 0, 33, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[3160, 103, 31, 0, 43, "Section"],
Cell[3194, 105, 102, 4, 33, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[3333, 114, 31, 0, 43, "Section"],
Cell[3367, 116, 9624, 187, 1097, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[13028, 308, 30, 0, 43, "Section"],
Cell[13061, 310, 412, 8, 71, "Text"],
Cell[13476, 320, 95, 2, 30, "Input",
  InitializationCell->True],
Cell[13574, 324, 221, 4, 70, "Input",
  InitializationCell->True],
Cell[13798, 330, 200, 4, 70, "Input",
  InitializationCell->True],
Cell[14001, 336, 310, 5, 90, "Input",
  InitializationCell->True],
Cell[14314, 343, 310, 5, 90, "Input",
  InitializationCell->True],
Cell[14627, 350, 674, 10, 190, "Input",
  InitializationCell->True],
Cell[15304, 362, 672, 11, 190, "Input",
  InitializationCell->True],
Cell[15979, 375, 341, 7, 90, "Input",
  InitializationCell->True],
Cell[16323, 384, 209, 4, 70, "Input",
  InitializationCell->True],
Cell[16535, 390, 214, 4, 70, "Input",
  InitializationCell->True],
Cell[16752, 396, 190, 4, 50, "Input",
  InitializationCell->True],
Cell[16945, 402, 201, 4, 70, "Input",
  InitializationCell->True],
Cell[17149, 408, 213, 5, 70, "Input",
  InitializationCell->True],
Cell[17365, 415, 256, 6, 70, "Input",
  InitializationCell->True],
Cell[17624, 423, 263, 5, 70, "Input",
  InitializationCell->True],
Cell[17890, 430, 261, 5, 70, "Input",
  InitializationCell->True],
Cell[18154, 437, 300, 5, 90, "Input",
  InitializationCell->True],
Cell[18457, 444, 329, 6, 90, "Input",
  InitializationCell->True],
Cell[18789, 452, 333, 6, 90, "Input",
  InitializationCell->True],
Cell[19125, 460, 268, 6, 70, "Input",
  InitializationCell->True],
Cell[19396, 468, 270, 5, 70, "Input",
  InitializationCell->True],
Cell[19669, 475, 260, 5, 70, "Input",
  InitializationCell->True],
Cell[19932, 482, 446, 8, 130, "Input",
  InitializationCell->True],
Cell[20381, 492, 450, 9, 130, "Input",
  InitializationCell->True],
Cell[20834, 503, 427, 7, 110, "Input",
  InitializationCell->True],
Cell[21264, 512, 438, 10, 110, "Input",
  InitializationCell->True],
Cell[21705, 524, 384, 6, 110, "Input",
  InitializationCell->True],
Cell[22092, 532, 167, 3, 50, "Input",
  InitializationCell->True],
Cell[22262, 537, 162, 3, 50, "Input",
  InitializationCell->True],
Cell[22427, 542, 162, 3, 50, "Input",
  InitializationCell->True],
Cell[22592, 547, 156, 3, 50, "Input",
  InitializationCell->True],
Cell[22751, 552, 159, 3, 50, "Input",
  InitializationCell->True],
Cell[22913, 557, 150, 3, 50, "Input",
  InitializationCell->True],
Cell[23066, 562, 176, 3, 50, "Input",
  InitializationCell->True],
Cell[23245, 567, 152, 3, 50, "Input",
  InitializationCell->True],
Cell[23400, 572, 178, 3, 50, "Input",
  InitializationCell->True],
Cell[23581, 577, 148, 3, 50, "Input",
  InitializationCell->True],
Cell[23732, 582, 158, 3, 50, "Input",
  InitializationCell->True],
Cell[23893, 587, 149, 3, 50, "Input",
  InitializationCell->True],
Cell[24045, 592, 187, 3, 50, "Input",
  InitializationCell->True],
Cell[24235, 597, 155, 3, 50, "Input",
  InitializationCell->True],
Cell[24393, 602, 155, 3, 50, "Input",
  InitializationCell->True],
Cell[24551, 607, 148, 3, 50, "Input",
  InitializationCell->True],
Cell[24702, 612, 88, 2, 30, "Input",
  InitializationCell->True]
}, Closed]],

Cell[CellGroupData[{
Cell[24827, 619, 35, 0, 43, "Section"],
Cell[24865, 621, 2075, 49, 223, "Text"],
Cell[26943, 672, 117, 5, 66, "Input",
  InitializationCell->True],
Cell[27063, 679, 1515, 34, 166, "Text"],
Cell[28581, 715, 148, 3, 30, "Input",
  InitializationCell->True],
Cell[28732, 720, 147, 3, 30, "Input",
  InitializationCell->True],
Cell[28882, 725, 720, 15, 166, "Text"],
Cell[29605, 742, 6442, 150, 774, "Text"],
Cell[36050, 894, 207, 4, 50, "Input",
  InitializationCell->True],
Cell[36260, 900, 812, 12, 128, "Text"],
Cell[37075, 914, 502, 10, 90, "Input",
  InitializationCell->True],
Cell[37580, 926, 4274, 110, 432, "Text"],
Cell[41857, 1038, 267, 6, 50, "Input",
  InitializationCell->True]
}, Closed]],

Cell[CellGroupData[{
Cell[42161, 1049, 66, 0, 43, "Section"],
Cell[42230, 1051, 1066, 24, 128, "Text"],
Cell[43299, 1077, 295, 6, 70, "Input",
  InitializationCell->True],
Cell[43597, 1085, 186, 6, 33, "Text"],
Cell[43786, 1093, 303, 6, 70, "Input",
  InitializationCell->True],
Cell[44092, 1101, 3252, 85, 394, "Text"],
Cell[47347, 1188, 289, 6, 70, "Input",
  InitializationCell->True],
Cell[47639, 1196, 1193, 30, 128, "Text"],
Cell[48835, 1228, 209, 5, 30, "Input",
  InitializationCell->True],
Cell[49047, 1235, 255, 8, 33, "Text"],
Cell[49305, 1245, 117, 2, 30, "Input",
  InitializationCell->True],
Cell[49425, 1249, 62, 0, 33, "Text"],
Cell[49490, 1251, 123, 3, 30, "Input",
  InitializationCell->True],
Cell[49616, 1256, 246, 6, 52, "Text"],
Cell[49865, 1264, 80, 2, 30, "Input",
  InitializationCell->True],
Cell[49948, 1268, 225, 4, 52, "Text"],
Cell[50176, 1274, 244, 5, 50, "Input",
  InitializationCell->True],
Cell[50423, 1281, 268, 9, 33, "Text"],
Cell[50694, 1292, 117, 2, 30, "Input",
  InitializationCell->True]
}, Closed]],

Cell[CellGroupData[{
Cell[50848, 1299, 38, 0, 43, "Section"],
Cell[50889, 1301, 1766, 40, 204, "Text"],

Cell[CellGroupData[{
Cell[52680, 1345, 46, 0, 38, "Subsection"],
Cell[52729, 1347, 522, 15, 71, "Text"],
Cell[53254, 1364, 187, 4, 50, "Input",
  InitializationCell->True],
Cell[53444, 1370, 133, 5, 33, "Text"],
Cell[53580, 1377, 194, 4, 50, "Input",
  InitializationCell->True]
}, Closed]],

Cell[CellGroupData[{
Cell[53811, 1386, 47, 0, 30, "Subsection"],
Cell[53861, 1388, 423, 10, 71, "Text"],
Cell[54287, 1400, 199, 5, 30, "Input",
  InitializationCell->True],
Cell[54489, 1407, 115, 2, 30, "Input",
  InitializationCell->True],
Cell[54607, 1411, 98, 2, 33, "Text"],
Cell[54708, 1415, 201, 5, 30, "Input",
  InitializationCell->True],
Cell[54912, 1422, 119, 2, 30, "Input",
  InitializationCell->True]
}, Closed]],

Cell[CellGroupData[{
Cell[55068, 1429, 48, 0, 30, "Subsection"],
Cell[55119, 1431, 979, 27, 109, "Text"],
Cell[56101, 1460, 143, 3, 30, "Input",
  InitializationCell->True],
Cell[56247, 1465, 147, 3, 30, "Input",
  InitializationCell->True]
}, Closed]],

Cell[CellGroupData[{
Cell[56431, 1473, 50, 0, 30, "Subsection"],
Cell[56484, 1475, 1408, 42, 128, "Text"],
Cell[57895, 1519, 798, 15, 190, "Input",
  InitializationCell->True]
}, Closed]],

Cell[CellGroupData[{
Cell[58730, 1539, 33, 0, 30, "Subsection"],
Cell[58766, 1541, 324, 11, 52, "Text"],
Cell[59093, 1554, 138, 3, 30, "Input",
  InitializationCell->True]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[59280, 1563, 54, 0, 43, "Section"],
Cell[59337, 1565, 621, 9, 109, "Text"],
Cell[59961, 1576, 162, 3, 30, "Input",
  InitializationCell->True],
Cell[60126, 1581, 142, 3, 30, "Input",
  InitializationCell->True],
Cell[60271, 1586, 293, 7, 52, "Text"],
Cell[60567, 1595, 189, 4, 50, "Input",
  InitializationCell->True],
Cell[60759, 1601, 630, 14, 90, "Text"],
Cell[61392, 1617, 572, 11, 130, "Input",
  InitializationCell->True],
Cell[61967, 1630, 133, 3, 33, "Text"],
Cell[62103, 1635, 214, 5, 50, "Input",
  InitializationCell->True],
Cell[62320, 1642, 605, 12, 130, "Input",
  InitializationCell->True],
Cell[62928, 1656, 1115, 25, 147, "Text"],
Cell[64046, 1683, 309, 6, 70, "Input",
  InitializationCell->True],
Cell[64358, 1691, 75, 0, 33, "Text"],
Cell[64436, 1693, 317, 7, 70, "Input",
  InitializationCell->True],
Cell[64756, 1702, 209, 4, 52, "Text"],
Cell[64968, 1708, 311, 6, 70, "Input",
  InitializationCell->True],
Cell[65282, 1716, 317, 7, 70, "Input",
  InitializationCell->True],
Cell[65602, 1725, 194, 4, 52, "Text"],
Cell[65799, 1731, 479, 9, 110, "Input",
  InitializationCell->True],
Cell[66281, 1742, 456, 9, 110, "Input",
  InitializationCell->True]
}, Closed]],

Cell[CellGroupData[{
Cell[66774, 1756, 45, 0, 43, "Section"],
Cell[66822, 1758, 584, 9, 90, "Text"],
Cell[67409, 1769, 340, 7, 70, "Input",
  InitializationCell->True],
Cell[67752, 1778, 593, 9, 90, "Text"],
Cell[68348, 1789, 253, 5, 50, "Input",
  InitializationCell->True],
Cell[68604, 1796, 1694, 31, 223, "Text"],
Cell[70301, 1829, 321, 6, 70, "Input",
  InitializationCell->True],
Cell[70625, 1837, 737, 21, 90, "Text"],
Cell[71365, 1860, 495, 8, 110, "Input",
  InitializationCell->True],
Cell[71863, 1870, 409, 7, 71, "Text"],
Cell[72275, 1879, 464, 9, 110, "Input",
  InitializationCell->True],
Cell[72742, 1890, 867, 15, 128, "Text"],
Cell[73612, 1907, 255, 5, 50, "Input",
  InitializationCell->True],
Cell[73870, 1914, 399, 8, 52, "Text"],
Cell[74272, 1924, 399, 8, 70, "Input",
  InitializationCell->True],
Cell[74674, 1934, 297, 5, 52, "Text"],
Cell[74974, 1941, 593, 12, 190, "Input",
  InitializationCell->True],
Cell[75570, 1955, 2596, 49, 318, "Text"],
Cell[78169, 2006, 599, 10, 130, "Input",
  InitializationCell->True],
Cell[78771, 2018, 414, 8, 71, "Text"],
Cell[79188, 2028, 790, 14, 150, "Input",
  InitializationCell->True],
Cell[79981, 2044, 587, 11, 90, "Text"],
Cell[80571, 2057, 831, 16, 250, "Input",
  InitializationCell->True],
Cell[81405, 2075, 1191, 23, 147, "Text"],
Cell[82599, 2100, 903, 15, 190, "Input",
  InitializationCell->True],
Cell[83505, 2117, 118, 2, 30, "Input",
  InitializationCell->True],
Cell[83626, 2121, 209, 6, 33, "Text"],
Cell[83838, 2129, 1093, 19, 190, "Input",
  InitializationCell->True],
Cell[84934, 2150, 118, 2, 30, "Input",
  InitializationCell->True],
Cell[85055, 2154, 740, 17, 90, "Text"],
Cell[85798, 2173, 288, 5, 90, "Input",
  InitializationCell->True],
Cell[86089, 2180, 208, 4, 50, "Input",
  InitializationCell->True],
Cell[86300, 2186, 946, 17, 128, "Text"],
Cell[87249, 2205, 292, 5, 50, "Input",
  InitializationCell->True],
Cell[87544, 2212, 66, 0, 33, "Text"],
Cell[87613, 2214, 308, 7, 50, "Input",
  InitializationCell->True],
Cell[87924, 2223, 99, 2, 33, "Text"],
Cell[88026, 2227, 233, 5, 30, "Input",
  InitializationCell->True],
Cell[88262, 2234, 210, 4, 52, "Text"],
Cell[88475, 2240, 209, 4, 50, "Input",
  InitializationCell->True],
Cell[88687, 2246, 581, 11, 90, "Text"],
Cell[89271, 2259, 424, 8, 70, "Input",
  InitializationCell->True],
Cell[89698, 2269, 298, 5, 52, "Text"],
Cell[89999, 2276, 396, 7, 130, "Input",
  InitializationCell->True],
Cell[90398, 2285, 306, 5, 52, "Text"],
Cell[90707, 2292, 192, 4, 50, "Input",
  InitializationCell->True],
Cell[90902, 2298, 105, 3, 33, "Text"],
Cell[91010, 2303, 249, 5, 50, "Input",
  InitializationCell->True],
Cell[91262, 2310, 2691, 47, 356, "Text"],
Cell[93956, 2359, 530, 10, 110, "Input",
  InitializationCell->True],
Cell[94489, 2371, 1786, 39, 204, "Text"],
Cell[96278, 2412, 416, 7, 110, "Input",
  InitializationCell->True],
Cell[96697, 2421, 660, 12, 90, "Text"],
Cell[97360, 2435, 538, 10, 130, "Input",
  InitializationCell->True],
Cell[97901, 2447, 5671, 176, 413, "Text"],
Cell[103575, 2625, 206, 4, 50, "Input",
  InitializationCell->True],
Cell[103784, 2631, 208, 4, 50, "Input",
  InitializationCell->True],
Cell[103995, 2637, 601, 22, 52, "Text"],
Cell[104599, 2661, 230, 5, 50, "Input",
  InitializationCell->True],
Cell[104832, 2668, 1655, 46, 147, "Text"],
Cell[106490, 2716, 252, 5, 50, "Input",
  InitializationCell->True],
Cell[106745, 2723, 444, 7, 90, "Text"],
Cell[107192, 2732, 940, 16, 170, "Input",
  InitializationCell->True],
Cell[108135, 2750, 413, 6, 71, "Text"],
Cell[108551, 2758, 496, 9, 110, "Input",
  InitializationCell->True],
Cell[109050, 2769, 88, 2, 33, "Text"],
Cell[109141, 2773, 476, 9, 110, "Input",
  InitializationCell->True],
Cell[109620, 2784, 1205, 31, 128, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[110862, 2820, 29, 0, 43, "Section"],
Cell[110894, 2822, 691, 16, 90, "Text"],
Cell[111588, 2840, 1172, 22, 350, "Input",
  InitializationCell->True],
Cell[112763, 2864, 1232, 26, 166, "Text"],
Cell[113998, 2892, 620, 13, 150, "Input",
  InitializationCell->True],
Cell[114621, 2907, 152, 3, 33, "Text"],
Cell[114776, 2912, 555, 11, 130, "Input",
  InitializationCell->True],
Cell[115334, 2925, 1147, 22, 147, "Text"],
Cell[116484, 2949, 846, 15, 170, "Input",
  InitializationCell->True],
Cell[117333, 2966, 88, 2, 33, "Text"],
Cell[117424, 2970, 850, 15, 190, "Input",
  InitializationCell->True],
Cell[118277, 2987, 1257, 22, 166, "Text"],
Cell[119537, 3011, 1067, 19, 230, "Input",
  InitializationCell->True],
Cell[120607, 3032, 1140, 24, 147, "Text"],
Cell[121750, 3058, 964, 18, 210, "Input",
  InitializationCell->True],
Cell[122717, 3078, 946, 21, 90, "Text"],
Cell[123666, 3101, 491, 9, 110, "Input",
  InitializationCell->True]
}, Closed]],

Cell[CellGroupData[{
Cell[124194, 3115, 29, 0, 43, "Section"],
Cell[124226, 3117, 3653, 74, 413, "Text",
  CellTags->"Refs"]
}, Closed]],

Cell[CellGroupData[{
Cell[127916, 3196, 75, 2, 43, "Section",
  Evaluatable->False],
Cell[127994, 3200, 79, 2, 30, "Input",
  InitializationCell->True],
Cell[128076, 3204, 86, 2, 30, "Input",
  InitializationCell->True]
}, Closed]]
}, Closed]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

